/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_base.cpp:  CBOLD fundamental classes

#include "cb_base.h"
#include "cb_prop.h"

                                              // string used for auto-named subbuses
                                              // user must not alter this variable, i.e., user must treat it as though it were const
char*        AutoSubBusString = "SB__";       // user must not use this string at the beginning of a connection name


// utilities

// these call ...fstream::open() -- if the file cannot be opened, they use BEGERR/ENDERR to output an error message via an exception
void OpenFileI( ifstream& Stream, const char* Path ) {
  Stream.open( Path, ios_base::in );
  if ( Stream.fail() )
    BEGERR << "Could not open file " << Path << " for input" << ENDERR;
}

void OpenFileO( ofstream& Stream, const char* Path ) {
  Stream.open( Path, ios_base::out );
  if ( Stream.fail() )
    BEGERR << "Could not open file " << Path << " for output" << ENDERR;
}

char* PermString( const char* CP ) {   // return a pointer to a permanent (non-local) copy of string pointed to by CP
  static TCharBucket CharBucket;       // holder of character strings created at run time
  return CharBucket.AddString( CP );   // pointer to permanent null-terminated char array
}

char* string_to_cstr( ostream& OS ) {
  ostringstream& OSS = *dynamic_cast<ostringstream*>( &OS );
  static TCharBucket CharBucket;                     // holder of character strings created at run time
  return CharBucket.AddString( OSS.str().c_str() );  // pointer to permanent null-terminated char array
}

const char* string_to_scstr( ostream& OS ) {
  ostringstream& OSS = *dynamic_cast<ostringstream*>( &OS );
  static string    S;
  S = OSS.str();
  return  S.c_str();  // pointer to static null-terminated char array
}


// define static members

TPortRange   TModule::DefaultRange;
TModule*     TModule::ConnectingModule     = 0;  // module whose Connect()     is currently being called--see shorthand operators
TModPar*     TModPar::RegisteringModPar    = 0;  // ModPar whose Register()    is currently being called
TModPar*     TModPar::PostConnectingModPar = 0;  // ModPar whose PostConnect() is currently being called

// Reg() for highest-level module
void  Reg( TRootModule* RootModule, char* aName, int aNumber ) {
  RootModule->Name.SetDesignator( aName, aNumber );            // set designator, but Owner will be 0
  RootModule->SerialNumber = RootModule->GetSerialNumber();    // always 0 for root module
  TModPar::RegisteringModPar = RootModule;
  RootModule->Register();                                      // register members
  TModPar::RegisteringModPar = 0;

#ifdef DEBUG_ON
    cout << "\nRegistered " << RootModule->Name << " at root level.\n";
#endif
}


// For null CBB, return null.
// For CBB derived from TModule or TPort, return owner.
// For CBB derived from TBundle, ascend owners until non-TBundle is found. Return this TModPar.
TModPar* GetUltimateOwner( TCBBase* CBB ) {
  TModPar* Owner = CBB->GetOwner();                       // get the immediate owner
  while ( 1 ) {                                           // find first non-TBundle owner
    if ( ! dynamic_cast<TBundle*>( Owner ) ) break;       // break when Owner is not a TBundle
    Owner = Owner->GetOwner();                            // get Owner's owner
  }
  return Owner;
}

// check that S is a valid name fragment, i.e., a string that contains only alphanumeric characters and underscore
void IsValidNameFragment( char* S, bool IgnoreEmpty ) throw ( TE_NameFragmentNull, TE_NameFragmentEmpty, TE_NameFragmentChar, TE_NameFragmentLength ) {
  if ( S == 0 ) {                                      // if null string
    throw TE_NameFragmentNull();                       // null strings are never ok
  }

  if ( S[ 0 ] == 0 ) {                                 // if empty string
    if ( IgnoreEmpty )  return;                        // empty string is ok
    else                throw TE_NameFragmentEmpty();  // empty string is not ok
  }

/*  locale loc;  // used by is...() */
  int i;
  for ( i = 0; i < 256; ++ i ) {
    char c = S[ i ];
    if ( c == 0 )  break;
    if ( isalnum(  c /*, loc */ ) )  continue;
    if ( c == '_' )                 continue;
    if ( !isprint( c /*, loc */ ) )  c = ' ';   // assume exception handler will want to show char to the user --> always return something printable
    throw TE_NameFragmentChar( c );
  }
  if ( i == 0  ||  i == 256 )  throw TE_NameFragmentLength( i );
}


ostream& operator<<( ostream& Dest, TDesignator D ) {

  TPath Path = TPath( D.Owner );

  for ( unsigned i = 1; i < Path.ModPars.size(); ++ i ) {  // for each of D's ancestors, oldest first, excluding root
    TModPar*   Ancestor  = Path.ModPars[ i ];

    char* B = Ancestor->Name.GetBase();    if ( B )        Dest << B;
    Dest  <<  Ancestor->Name.GetString();
    int   N = Ancestor->Name.GetNumber();  if ( N != -1 )  Dest << N;
    Dest  << '.';
  }

  char* B = D.GetBase();    if ( B )        Dest << B;
  Dest  <<  D.GetString();
  int   N = D.GetNumber();  if ( N != -1 )  Dest << N;

  return Dest;
}

// like operator<<(), but do not output full path
ostream& Out_D_NP(  ostream& Dest, TDesignator D ) {
  char* B = D.GetBase();    if ( B )        Dest << B;
  Dest  <<  D.GetString();
  int   N = D.GetNumber();  if ( N != -1 )  Dest << N;

  return Dest;
}


ostream& operator<<( ostream& Dest, TCBBase CBB ) {
  return Dest << CBB.Name;
}



ostream& operator<<( ostream& Dest, TPortRange PR ) {
  Dest << PR.GetPort()->Name;
  if ( ( PR.GetPort()->GetCount() != 1 ) || ( PR.GetMinRange() != 0 ) ) {  // do not output range details if port is zero-based single-wire port
    Dest << '[' << PR.GetLeftRange();
    if ( PR.GetLeftRange() != PR.GetRightRange() ) Dest << ':' << PR.GetRightRange();
    Dest << ']';
  }
  return Dest;
}


// like operator<<(), but do not output full path
ostream& Out_PR_NP(  ostream& Dest, TPortRange   PR  ) {
  Dest << PR.GetPort()->Name.GetString();
  if ( ( PR.GetPort()->GetCount() != 1 ) || ( PR.GetMinRange() != 0 ) ) {  // do not output range details if port is zero-based single-wire port
    Dest << '[' << PR.GetLeftRange();
    if ( PR.GetLeftRange() != PR.GetRightRange() ) Dest << ':' << PR.GetRightRange();
    Dest << ']';
  }
  return Dest;
}


// output port name if R is really a TPort or TPortRange, else just output range
ostream& operator<<( ostream& Dest, TRange* R ) {
  if ( ! R )  return Dest << "*** NULL TRange ***";
  TPortRange* PR = dynamic_cast<TPortRange*>( R );
  if ( PR )  return Dest << *PR;                  // if R is base of a TPortRange

  TPort* P = dynamic_cast<TPort*>( R );
  if ( P )  return Dest << *P;                    // if R is base of a TPort

  Dest << '[' << R->GetLeftRange();
    if ( R->GetLeftRange() != R->GetRightRange() ) Dest << ':' << R->GetRightRange();
  Dest << ']';
  return Dest;
}

// output module name if SBL is really TModule*, else output unknown
ostream& operator<<( ostream& Dest, TSubBusList* SBL ) {
  if ( !SBL )  return Dest << "*** NULL TSubBusList ***";
  TModule* M = dynamic_cast<TModule*>( SBL );
  if ( M )  return Dest << "SubBusList of " << M;
  return Dest << "TSubBusList";               // unlikely, because TSubBusList's are normally only used as base of TModule
}

// output module name
ostream& operator<<( ostream& Dest, TModule* M ) {
  if ( !M )  return Dest << "*** NULL TModule ***";
  return Dest << M->Name;
}


ostream& operator<<( ostream& Dest, TPinSpec PS ) {
  TPort* Port = PS.GetPort();
  int Index   = PS.GetIndex();

  TPart* OP = dynamic_cast<TPart*>( Port->GetOwner() );  // if owner is a part, output reference designator, etc.
  if ( OP ) {
    size_t PinCount = 1;
    if ( Port->MultiPinNode() ) { Index = 0; PinCount = Port->GetPinCount(); }
    for ( size_t i = 0; i < PinCount; ++ i ) {
      Dest << ' ';
      Dest << OP->GetReferenceDesignator();
      Dest << '.' << Port->GetPin( Index++ );
    }
  }
  else  Dest << Port;        // otherwise output port symbolic name

  return Dest;
}

ostream& operator<<( ostream& Dest, TPackage P ) {
  Dest << P.Name;
  return Dest;
}


ostream& operator<<( ostream& Dest, TNetName NN ) {
  return Dest << &NN;
}

ostream& operator<<( ostream& Dest, TNetName* NN ) {
  if ( !NN )  return Dest << "*** NULL TNetName ***";
  Dest << TDesignator ( *NN );
  if ( NN->GetExtension() != 0 )  Dest << "__" << NN->GetExtension();
  return Dest;
}

void operator<<( TLogStatus& L, char* aString ) {
  L.Log( aString );
};


void MessageIfError( ostream& Dest, TCBBase* ErrorSource, char* PassName ) {   // used only in main()'s exception handler
  if ( !ErrorSource )  return;
  Dest << "\nError detected while " << ErrorSource->GetTypeName() << " " << *ErrorSource << " was " << PassName << ".\n";
}

bool operator<( TDesignator A, TDesignator B ) {  // use with stdlib's sort()
  if ( A.GetBase() != B.GetBase() ) {
    if ( A.GetBase() == 0 )  return true;         // designators with null Base come first after sort()
    if ( B.GetBase() == 0 )  return false;
    int Cmp = strcmp( A.GetBase(), B.GetBase() );
    if ( Cmp != 0 )  return Cmp < 0;
  }

  int Cmp = strcmp( A.GetString(), B.GetString() );
  if ( Cmp != 0 )  return Cmp < 0;
  return A.GetNumber() < B.GetNumber();
}

bool operator<( TNetName A, TNetName B ) {            // use with stdlib's sort()
  TDesignator AD = A;
  TDesignator BD = B;                                 // compare as designators
  if ( AD <  BD )  return true;                       // if A's designator is < B's designator
  if ( ! AD.DesignatorEquals( BD ) )  return false;   // if A's designator is > B's designator
  return A.GetExtension() < B.GetExtension();         // if designators are identical, then use extension (e.g., __1, __2)
}

bool operator<( TPinSpec A, TPinSpec B ) {      // use with stdlib's sort()
  int A_SN = A.GetPort()->SerialNumber;         // first consider registration order of ports
  int B_SN = B.GetPort()->SerialNumber;
  if ( A_SN != B_SN )  return A_SN < B_SN;      // if different ports, first registered is less
  return A.GetIndex() < B.GetIndex();           // if same port, lowest index is less
}


void CBOLD_CopyError( char* ClassName, char* Mode) {
  BEGERR << "Copying (by " << Mode << ") of objects of type " << ClassName << " is illegal. The desired functionality might be achieved by using a reference or pointer instead" << ENDERR;
}


bool TPortRange::Registered() const { return Port->Name.Registered(); }


// merge all wires of Mergee into single wire of Other
void TModule::merge( TPortRange Other, TPortRange Mergee ) {
  if ( Other.GetCount() != 1 )
    BEGERR << "merge( " << Other << ", " << Mergee << " ):  first argument must have width 1" << ENDERR;
  for ( int i = Mergee.GetMinRange(); i <= Mergee.GetMaxRange(); ++ i ) {
    conn( Other, (*Mergee.GetPort())( i ) );
  }
}

// merge all wires of Mergee into single wire of aNamed
void TModule::merge( char*     aNamed, TPortRange Mergee ) {
  for ( int i = Mergee.GetMinRange(); i <= Mergee.GetMaxRange(); ++ i ) {
    conn( aNamed, (*Mergee.GetPort())( i ) );
  }
}


// conn() for connecting array of ports to named nets

// each port in the array is connected to a separate net
// the net names consist of the following:
//    prefix       // user-specified prefix
//    index        // index of port within array
//    underscore   // underscore (only if port has more than one wire)
// the actual character strings are maintained in a permanent string container

// it is legal for Prefix to begin with GlobalChar


char* TModule::conn( char* Prefix, TPort* Ports, int PortCount ) {
  string NetPrefix = Prefix;

  for ( int i = 0; i < PortCount; ++ i ) {
    TPort* Port = Ports + i;
    string NetSuffix =  SCSTRING( Port->Name.GetNumber() );    // convert number to string
    if ( Port->GetCount() > 1 )  NetSuffix += "_";             // if multi-wire port, separate array index from wire index

    string NetName    = NetPrefix + NetSuffix;
    char*  NetNamePtr = PermString( NetName.c_str() );         // pointer to permanent null-terminated char array

    conn( NetNamePtr, *Port );
  }

  return Prefix;
}






TSubBus* TModule::conn( char* aName, const TPortRange& aConn0,
                                     const TPortRange& aConn1,
                                     const TPortRange& aConn2,
                                     const TPortRange& aConn3,
                                     const TPortRange& aConn4,
                                     const TPortRange& aConn5,
                                     const TPortRange& aConn6,
                                     const TPortRange& aConn7,
                                     const TPortRange& aConn8,
                                     const TPortRange& aConn9 )
{
  TDesignator Designator;
  bool Global = ( this == GlobalModule() );
  if ( ! Global )                      // handle global names
    if ( *aName == GlobalChar )
      return( GlobalModule()->conn( aName, aConn0, aConn1, aConn2, aConn3, aConn4,
                                           aConn5, aConn6, aConn7, aConn8, aConn9 ));


// move arguments into array, check that they are legal
  TPortRange Conn[ 10 ];
  Conn[ 0 ] = aConn0;  Conn[ 1 ] = aConn1;
  Conn[ 2 ] = aConn2;  Conn[ 3 ] = aConn3;
  Conn[ 4 ] = aConn4;  Conn[ 5 ] = aConn5;
  Conn[ 6 ] = aConn6;  Conn[ 7 ] = aConn7;
  Conn[ 8 ] = aConn8;  Conn[ 9 ] = aConn9;
  for ( int i = 0; i < 10; ++ i ) {
    if ( ! Conn[ i ].Defined() )  break;    // watch for first default argument (if any)
    if ( ! Conn[ i ].Registered() )
      BEGERR << "conn(): port must be registered: " << Conn[ i ] << ENDERR;
    if ( ! Global )
      if ( ! PortAccessAllowed( Conn[ i ].GetPort() ) )
        BEGERR << "conn(): " << this << " is not allowed access to " << Conn[ i ] << ENDERR;
//    if ( Global )                                 // >>> this code is insufficient
//      if ( IsMember( Conn[ i ].GetPort() ) )
//        BEGERR << "conn():  module cannot connect its port to a global net" << ENDERR;

// this test removed 8-29-04
// it is now legal for two of a module's ports to share a common internal connection
//  if ( ! IsMember( Conn[ i ].GetPort() ) ) {   // if connecting something besides one of this module's ports
//    TPortRange* PR = MemberPortOverlap( Conn[ i ], this );
//    if ( PR )            // if SubBusList already contains some wires from this port range that are connected to one of this module's ports
//      BEGERR << "conn(): two of a module's ports cannot share a common internal connection. Port " << Conn[ i ] << " overlaps port " << *PR << ENDERR;
//  }
  }


// determine target SubBus
// Though it is legal for a connection to appear twice in the SubBusList, processing
// of the netlist can be greatly slowed by duplicate connections.
// The code below does not create duplicates except when the user connects
// the same port range to more than one user-named SubBus.  E.g.,:
//    conn( "foo", myport );  conn( "bar", myport );

  TSubBus* Target = 0;

  if ( *aName == 0 ) {                        // if user supplied no name
    for ( int i = 0; i < 10; ++ i ) {         // check if any of the new connections are already in this module's SubBusList
      if ( ! Conn[ i ].Defined() )  break;    // watch for first default argument (if any)
      for ( unsigned s = 0; s < SubBuses.size(); ++ s ) {
        if ( SubBuses[ s ].IsConnection( Conn[ i ] ) ) {      // if connection is in SubBuses[ s ]
          Target = & SubBuses[ s ];
          break;
        }
      }
      if ( Target ) break;
    }
    if ( Target == 0 ) {    // if no connection is in SubBusList, use automatically generated name
      Designator.SetDesignator( AutoSubBusString, AutoSubBusNumber++ );  // AutoSubBusNumber starts at 1
    }
  }
  else {                    // if user supplied a name
    Designator.SetDesignator( aName );
    for ( unsigned i = 0; i < SubBuses.size(); ++ i ) {  // if name already exists, add connection(s) to existing SubBus
      if ( SubBuses[ i ].DesignatorEquals( Designator ) ) {
        Target = & SubBuses[ i ];
        break;
      }
    }
  }


// if no existing Target was found, create it
  if ( Target == 0 ) {
    TSubBus Temp;
    Temp.SetDesignator( Designator );
    SubBuses.push_back( Temp );
    Target = &SubBuses.back();
  }

// make the connection
  for ( int i = 0; i < 10; ++ i ) {
    if ( ! Conn[ i ].Defined() )  break;    // watch for first default argument (if any)
    Target->Add( Conn[ i ] );               // add connection to target SubBus
#ifdef DEBUG_ON
    cout << "\nAdded connection " << Conn[ i ] << " to " << *Target << '.';
#endif
  }
  return Target;
};


// return a port range in the subbus -- allows operations on the subbus via its name, e.g., associating subbus with a property
TPortRange TModule::bus( char* SubBusName ) {
  bool Global = ( this == GlobalModule() );
  if ( ! Global )                      // handle global names
    if ( *SubBusName == GlobalChar )
      return( GlobalModule()->bus( SubBusName ) );

  for ( unsigned s = 0; s < SubBuses.size(); ++ s ) {          // search subbus list for named subbus
    TSubBus* SubBus = &SubBuses[ s ];
    if ( SubBus->DesignatorEquals( SubBusName ) ) {            // if name matches
      if ( SubBus->GetConnectionCount() == 0 )                 // should never happen?
        BEGERR << "TModule::bus( Name ): the named SubBus, " << SubBusName << ", has no connections" << ENDERR;
      return SubBus->GetConnection( 0 );                       // return first PortRange listed in SubBus
    }
  }
  BEGERR << "TModule::bus(): the named SubBus, " << SubBusName << ", does not exist" << ENDERR;
}

// create a new subbus using an internal port and the specified range
// SubBusName must not already exist in the SubBus list
TPortRange TModule::bus( char* SubBusName, int LeftRange, int RightRange ) {
  bool Global = ( this == GlobalModule() );
  if ( ! Global )                      // handle global names
    if ( *SubBusName == GlobalChar )
      return( GlobalModule()->bus( SubBusName, LeftRange, RightRange ) );

  for ( unsigned s = 0; s < SubBuses.size(); ++ s ) {          // verify that subbus list does not already contain SubBusName
    TSubBus* SubBus = &SubBuses[ s ];
    if ( SubBus->DesignatorEquals( SubBusName ) ) {            // if name matches
      BEGERR << "TModule::bus( Name, L, R ): the named SubBus, " << SubBusName << ", already exists. Perhaps you want bus( Name )( L, R ) instead" << ENDERR;
    }
  }

  TPort* P = NewInternalPort( SubBusName, LeftRange, RightRange );
  conn( SubBusName, *P );
  return *P;
}

// create new internal port
TPort* TModule::NewInternalPort( char* Name, int LeftRange, int RightRange ) {
  TPort* P = new TPort;
  if ( !P )  BEGERR << "Non-user error:  TModule::NewInternalPort() could not create a new TPort using operator new" << ENDERR;
  P->SetPortType( ptInternal );        // set type, because some code needs to know that this is an internal port
  InternalPorts.push_back( P );        // save pointer for delete in ~TModule

  Reg( P, CSTRING( "internal_bus__" << Name ) );   // register the port using a destinctive name, e.g., to appear in database dump, error messages, etc.
  P->SetRange( LeftRange, RightRange );
  return P;
}


int TModule::wire( TPortRange MyPortRange, char* MemberPortName, int MaxMemberPorts, int LeftRange, int RightRange ) {
  // check that port of MyPortRange is really a member of this module
//  if ( ! IsMember( MyPortRange.GetPort() ) )
  if ( GetUltimateOwner( MyPortRange.GetPort() )  !=  this )
    BEGERR << "wire( " << MyPortRange << ", ...): first argument must be a port of calling module" << ENDERR;

  if ( RightRange == -1 )     RightRange = LeftRange;                    // single wire
  if ( MemberPortName == 0 )  MemberPortName = MyPortRange.GetPort()->Name.GetString();  // default name for member ports

// create a new connection to port from this module
  TSubBus* SubBus = conn( MyPortRange );

// add port(s) of module/part (but not bundle) members of this module to the same connection
  int SuccessCount = 0;
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    if (          dynamic_cast< TBundle* >( Member ) != 0 )  continue;  // do not consider member bundles
    TModPar* MP = dynamic_cast< TModPar* >( Member );
    if ( MP == 0 )  continue;
    TPort* MemberPort = MP->FindPort( MemberPortName );
    if ( MemberPort == 0 )  continue;
    if ( LeftRange == -1 )  SubBus->Add(  *MemberPort                           );  // use full range
    else                    SubBus->Add( (*MemberPort)( LeftRange, RightRange ) );
    ++ SuccessCount;
  }
  if ( SuccessCount == 0 )
    BEGERR << "wire( " << MyPortRange << ", \"" << MemberPortName << "\" ) could not connect any ports of any members" << ENDERR;
  if ( MaxMemberPorts != -1 )
    if ( SuccessCount > MaxMemberPorts )
      BEGERR << "wire( " << MyPortRange << ", \"" << MemberPortName << "\" ): MaxMemberPorts ( " << MaxMemberPorts << " ) exceeded" << ENDERR;
  return SuccessCount;
}


int  TModule::wire( char* aNamed, char* MemberPortName, int MaxMemberPorts, int LeftRange, int RightRange ) {

  if ( RightRange     == -1 )  RightRange = LeftRange;   // single wire
  if ( MemberPortName ==  0 )  MemberPortName = aNamed;    // default name for member ports

// !!! there seems to be no reason to have this limitation except to protect the novice user--a warning might be appropriate
//  if ( FindPort( MemberPortName ) != 0 )
//    BEGERR << "wire( " << aNamed << ", " << MemberPortName << " ... ): " << MemberPortName << " is the name of one of this module's ports" << ENDERR;

// add port(s) of module/part (but not bundle) members of this module to the same connection
  int SuccessCount = 0;
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    if (          dynamic_cast< TBundle* >( Member ) != 0 )  continue;  // do not consider member bundles
    TModPar* MP = dynamic_cast< TModPar* >( Member );
    if ( MP == 0 )  continue;
    TPort* MemberPort = MP->FindPort( MemberPortName );
    if ( MemberPort == 0 )  continue;
    if ( LeftRange == -1 )  conn( aNamed,  *MemberPort                           );  // use full range
    else                    conn( aNamed, (*MemberPort)( LeftRange, RightRange ) );
    ++ SuccessCount;
  }
  if ( SuccessCount == 0 )
    BEGERR << "wire( " << aNamed << ", " << MemberPortName << " ... ): could not connect any ports of any members" << ENDERR;
  if ( MaxMemberPorts != -1 )
    if ( SuccessCount > MaxMemberPorts )
      BEGERR << "wire( " << aNamed << ", " << MemberPortName << " ... ): MaxMemberPorts ( " << MaxMemberPorts << " ) exceeded" << ENDERR;
  return SuccessCount;
}


// wire identically-named (and ranged) ports of Bundle and ModPar together

int TModule::wire( TBundle& Bundle, TModPar& ModPar ) {
  int S = wire_int( Bundle, ModPar );   // call recursive function that accumulates SuccessCount
  if ( S == 0 )                             // if no success (no connections)
    BEGERR << "wire( " << Bundle << ", " << ModPar << " ): could not connect any ports of " << ModPar << ENDERR;
  return S;
}

// formerly wire(), now a recursive helper function for wire()
int TModule::wire_int( TBundle& Bundle, TModPar& ModPar ) {
  int SuccessCount = 0;

  for ( TCBBase* Member = Bundle.GetFirstMember(); Member != 0; Member = Member->Next ) {  // for each member of Bundle
    TPort* Port         = dynamic_cast<TPort*>(   Member );
    TBundle* SubBundle  = dynamic_cast<TBundle*>( Member );
    if ( Port ) {
      SuccessCount += wire_int( Port, ModPar );    // check for ModPar's ports that match Port, descend recursively into ModPar's bundles
    }
    else if ( SubBundle ) {
      SuccessCount += wire_int( *SubBundle, ModPar );  // descend recursively into Bundle's sub-bundles
    }
    else
      BEGERR << "wire( " << Bundle << ", " << ModPar << " ): Unexpected type encountered as member of " << Bundle << ENDERR;
  }

  return SuccessCount;
}

int TModule::wire_int( TPort* Port, TModPar& ModPar ) {
  int SuccessCount = 0;

  for ( TCBBase* MP_Member = ModPar.GetFirstMember(); MP_Member != 0; MP_Member = MP_Member->Next ) {
    TBundle* MP_Bundle = dynamic_cast<TBundle*>( MP_Member );
    if ( MP_Bundle ) {
      SuccessCount += wire_int( Port, *MP_Bundle );  // descend recursively into ModPar's bundle
      continue;
    }
    TPort*   MP_Port   = dynamic_cast<TPort*>( MP_Member );

    if ( ! MP_Port )  continue;                                                        // if not port
    if ( strcmp( Port->Name.GetString(), MP_Port->Name.GetString() ) )  continue;      // if name does not match
    if ( Port->GetLeftRange()  !=  MP_Port->GetLeftRange()  )  continue;               // if range does not match
    if ( Port->GetRightRange() !=  MP_Port->GetRightRange() )  continue;
    conn( *Port, *MP_Port );
    SuccessCount ++;
  }

  return SuccessCount;
}



void TModPar::Reg( TCBBase* aMember, char* aName, int aNumber ) {   // register a member with the module
  if ( aMember->GetOwner() != 0 )
    BEGERR << "Attempt to register member twice: " << *aMember << ENDERR;
  aMember->Name.SetDesignator( aName, aNumber, this );

  TModPar*   MP = dynamic_cast<TModPar*>( aMember );
  if ( MP )  MP->Root = Root;

  aMember->SerialNumber = Root->GetSerialNumber();  // set serial (registration) number

  aMember->Prev    = LastMember;                    // Prev == 0 if first member
  aMember->Next    = 0;                             // Next == 0 if last  member

  if ( FirstMember == 0 )  FirstMember      = aMember;
  if ( LastMember  != 0 )  LastMember->Next = aMember;
  LastMember                                = aMember;
  MemberCount++;
  if ( dynamic_cast<TPort*>( aMember ) )  PortCount ++;  // very little code relies on PortCount

  RegisteringModPar = MP;
  if ( MP ) MP->Register();
  RegisteringModPar = 0;

#ifdef DEBUG_ON
  cout << "\nRegistered " << aMember->Name << " with " << Name << '.';
#endif
}

inline TModule*  TModPar::GlobalModule() { return & Root->GlobalSubBusModule; }



TPort* TModPar::FindPort( char* aString, int aNumber ) {    // find port with specified designator
  if ( dynamic_cast< TBundle* >( this ) != 0 )
    BEGERR << "Call to TModPar::FindPort() not allowed for TBundle (" << *this << "). This is an internal CBOLD error" << ENDERR;     //
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    TPort* P = dynamic_cast< TPort* >( Member );
    if ( P == 0 )  continue;            // continue if member is not a port
    if ( strcmp( aString, P->Name.GetString() ) != 0 )  continue;
    if ( aNumber == P->Name.GetNumber() )     return P;
  }
  return 0;
}



// Return true if P is accessible to this module/part. Access is allowed if:
//   P is a port whose owner is:
//     this module, or
//     a member of this module, OR
//   P is a port of a TBundle or nested TBundle whose owner is:
//     this module, or
//     a member of this module
bool TModule::PortAccessAllowed( TPort* P ) {
  TModPar* UltimateOwner = GetUltimateOwner( P );
  if ( UltimateOwner == this )              return true;   // if owner is this module
  if ( UltimateOwner->GetOwner() == this )  return true;   // if owner is a member of this module
  return false;
}


TPort* TModPar::GetPort( int N ) {   // get the Nth port ( N = 0 for the first port registered, etc.)
  int CurrentPort = 0;
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    TPort* P = dynamic_cast< TPort* >( Member );
    if ( P ) {
      if ( CurrentPort++  ==  N )  return P;
    }
  }
  BEGERR << "TModPar::GetPort( " << N << " ): no such port" << ENDERR;
}


void  TModule::FillNetList( TNetList* NetList ) {

// The order in which SubBuses are added to the NetList is important:
//   1. Global nets
//   2. SubBuses of this module, in the order connected.
//   3. SubBuses of member modules, in the order the modules were registered.
// With this ordering, TNetList::Cleanup() can pick a reasonable net name for nets that have aliases.

// global SubBusList must be added first
// an error results if a PinSpec already in a global net is added to the netlist
  if ( NetList->Cleared() ) {
    NetList->AddSubBusList( *GlobalModule() );  // add global SubBuses to netlist
  }

// add any connections that originate in this module
// some connections in SubBusList may already be in netlist due to higher-level modules adding SubBuses to netlist
  NetList->AddSubBusList( *this );

// add any connections that originate in member modules
// some connections in member modules may already be in netlist due to higher-level modules adding SubBuses to netlist
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    TModPar* MP = dynamic_cast< TModPar* >( Member );
    if ( MP ) MP->FillNetList( NetList );
  }
}



// Add SubBus to netlist, one net per wire.
// Net names are derived from the SubBus name and are not guaranteed unique.
// Call Cleanup() to remove empty nets, to merge nets that share pins, and to assign extensions to duplicate net names.
void TNetList::AddSubBus( TSubBus& SubBus ) {
  mCleared = false;
  TRange   Range = SubBus.GetRange();       // Range of SubBus being added to netlist
  int  LeftRange = Range.GetLeftRange();
  int RightRange = Range.GetRightRange();

  // for each wire of the SubBus
  for ( int i = LeftRange; ; i += ( ( RightRange > LeftRange ) ? 1 : -1 ) ) {
    TNet* N = NewNet();
// >>>    if ( N->CheckGlobalNameConflict )
// >>>     BEGERR << "Warning: Local net name is same as global net name less '/'" << ENDERR;
    N->SetDesignator( SubBus );  // net's name is same as SubBus's name
    if ( LeftRange != RightRange )  N->ReplaceNumber( i ); // if not single wire

    // for each connection, add all of its physical pins to the net
    for ( unsigned c = 0; c < SubBus.GetConnectionCount(); ++ c ) {
      TPortRange  Connection = SubBus.GetConnection( c );
      int Index = Connection.TranslateIndex( Range, i );   // translate from index of SubBus to index of this connection
      TPort* Port = Connection.GetPort();
      N->AddPin( Port, Index );
    }
    if ( i == RightRange ) break;
  }
}

void TNetList::AddSubBusList( TSubBusList& SubBusList, char* ExcludeName ) {
  mCleared = false;
  for ( unsigned i = 0; i < SubBusList.GetSubBusCount(); ++ i ) {
    if ( ExcludeName[ 0 ]  !=  0 )   // for speed only
      if ( SubBusList.GetSubBus( i ).DesignatorEquals( ExcludeName ) ) continue;
    AddSubBus( SubBusList.GetSubBus( i ) );
  }
}

TModule::~TModule() { for ( unsigned i = 0; i < InternalPorts.size(); ++ i )  delete InternalPorts[ i ]; }
TPropertyList::~TPropertyList() {
  while ( ! List.empty() ) {
    delete List.back();
    List.pop_back();
  }
}


TPortRange* TSubBusList::PortOverlap( TPortRange aPortRange ){    // return overlapping port range, else 0
  for ( unsigned i = 0; i < GetSubBusCount(); ++ i ) {
    TPortRange* PR = SubBuses[ i ].PortOverlap( aPortRange );
    if ( PR )  return PR;
  }
  return 0;
}


TPortRange* TSubBus::PortOverlap( TPortRange aPortRange ){    // return overlapping port range if SubBus contains a connection that overlaps aPortRange
  for ( unsigned i = 0; i < GetConnectionCount(); ++ i ) {
    if ( Connections[ i ].Overlap( aPortRange ) ) return &Connections[ i ];
  }
  return 0;
}


// These are like the PortOverlap() functions, except a SubBus must contain a port of ModPar to be considered overlapping

TPortRange* TSubBusList::MemberPortOverlap( TPortRange aPortRange, TModPar* ModPar ){    // return overlapping port range, else 0
  for ( unsigned i = 0; i < GetSubBusCount(); ++ i ) {
    TPortRange* PR = SubBuses[ i ].MemberPortOverlap( aPortRange, ModPar );
    if ( PR )  return PR;
  }
  return 0;
}


TPortRange* TSubBus::MemberPortOverlap( TPortRange aPortRange, TModPar* ModPar ){    // return overlapping range if SubBus contains a connection that overlaps aPortRange
  for ( unsigned i = 0; i < GetConnectionCount(); ++ i ) {              // for each connection in the SubBus
    if ( ModPar->IsMember( Connections[ i ].GetPort() ) ) {             // if the connection is to a member port of ModPar
      for ( unsigned j = 0; j < GetConnectionCount(); ++ j ) {          // then check if aPortRange overlaps any connection in the SubBus
        if ( Connections[ j ].Overlap( aPortRange ) ) return &Connections[ j ];
      }
      return 0;   // ok to return, because the for loop immediately above is independent of ModPar
    }
  }
  return 0;
}


int TSubBus::FindPin( TPinSpec aPinSpec ){  // return index of wire in SubBus that is connected to aPinSpec, else -1
  int Index = aPinSpec.GetIndex();
  for ( unsigned i = 0; i < GetConnectionCount(); ++ i ) {
    TPortRange PortRange = Connections[ i ];
    if ( PortRange.GetPort() != aPinSpec.GetPort() )  continue;   // port must match
    if ( ! PortRange.InRange( Index ) )               continue;   // index must be within range
    return GetRange().TranslateIndex( PortRange, Index );         // translate from index within port to index within SubBus
  }
  return -1;
}



//Cleanup():
//  1. remove empty nets
//  2. merge nets that share pins
//  3. assign extensions to duplicate net names
void TNetList::Cleanup( bool Sort ) {
  size_t NetCount = Nets.size();

  if ( NetCount != 0 ) {
    for ( size_t i = NetCount-1; ; -- i ) {           // initialize each PinSpec's Port's net index to the net's index in Nets
      Nets[ i ].SetNetIndex( (unsigned)(i) );
      if ( i == 0 ) break;
    }
  }

  int  Passes = 0;
  bool NetIndexChanged = true;
  while ( NetIndexChanged ) {                                 // loop until no PinSpec's net index changed --> all connections have been resolved
    NetIndexChanged = false;
    for ( unsigned i = 0; i < NetCount; ++ i ) {              // set each PinSpec's Port's net index to the minimum of i or minimum net index in net
      if ( Nets[ i ].SetMinNetIndex( i ) ) {
        NetIndexChanged = true;               // do not use || else call to SetMinIndex() is short circuited
      }
    }
    Passes ++;
  };

  for ( unsigned i = 0; i < NetCount; ++ i ) {                // merge nets that have common PinSpecs
    TNet& SrceNet = Nets[ i ];
    int NetIndex = SrceNet.GetNetIndex();                     // get new (destination) net index
    if ( NetIndex  < 0 )  continue;                           // should never happen (empty net)
    if ( (unsigned) NetIndex != i ) {                         // if not already minimum net index
      TNet& DestNet = Nets[ unsigned( NetIndex ) ];
      if ( DestNet.DesignatorEquals( NoConnect )  &&  ( SrceNet.GetPartPinSpecCount() > 1 ) )
        BEGERR << "TNetList::Cleanup() found net " << SrceNet << " connected to the global no-connect net " << NoConnect << ENDERR;
      if ( ( DestNet.GetString() == AutoSubBusString )  &&     // the first user-assigned net name (if any) will be the name for the complete net
           ( SrceNet.GetString() != AutoSubBusString )     )  DestNet.ReplaceDesignator( SrceNet );
      DestNet.AddNet( &SrceNet );                              // move PinSpecs to net with lower index
      SrceNet.Clear();                                         // flag net as having been merged
    }
  }


  vector<TNet> NetList;     // final netlist is built here
  for ( unsigned i = 0; i < Nets.size(); ++ i ) {
    TNet& SrceNet = Nets[ i ];
    if ( SrceNet.GetPinSpecCount() == 0 ) continue;
    NetList.push_back( SrceNet );
  }

// assign extensions (>0) to duplicate net names
  if ( NetList.size() >= 2 ) {
    for ( unsigned s = 1; s < NetList.size(); ++ s ) {  // for each net except the first
      TNet& SrceNet = NetList[ s ];
      for ( unsigned p = s; p != 0; -- p ) {            // for all previous nets, in reverse order
        TNet& PrevNet = NetList[ p - 1 ];
        if ( SrceNet.DesignatorEquals( PrevNet ) ) {
          SrceNet.SetExtension( PrevNet.GetExtension() + 1 );
          break;
        }
      }
    }
  }

  Nets = NetList;
  
  if ( Sort ) {
    SortNets( false );    // sort nets alphabetically by name -- false --> do not call SetNetIndexes() -- for efficiency, SetNetIndexes() is called below
    SortPinSpecs();       // sort each net's PinSpecs by port registration order and index
  }

  SetNetIndexes();        // assign final net indexes

  mCleaned = true;        // true --> net indexes in PinSpec can be used to identify net containing PinSpec

#if 0
// for efficiency, copy vector Nets to list NetList
  list<TNet> NetList;
  NetList.assign( Nets.begin(), Nets.end() );

// remove empty nets
  for ( list<TNet>::iterator I = NetList.begin(); I != NetList.end(); ) {
    if ( I->PinSpecs.size() == 0 )  I = NetList.erase( I );
    else                            I++;
  }


// Pass 1:  incomplete pass merges most nets efficiently
//  for ( bool AnyMerged = true;  AnyMerged;  ) {  // repeat until no merges occur
//    AnyMerged = false;
  // merge nets that share pins
    for ( list<TNet>::iterator I = NetList.begin(); I != NetList.end(); ++ I ) {
      list<TNet>::iterator J = I; ++ J;   // operator+() not defined for list iterators
      while ( J != NetList.end() ) {
        if ( I->CommonPins( J.operator->() ) ) {
          if ( I->DesignatorEquals( NoConnect )  &&  J->GetPartPinSpecCount() > 1 )
            BEGERR << "TNetList::Cleanup() found net " << J.operator->() << " connected to the global no-connect net " << NoConnect << ENDERR;
          if ( ( I->GetString() == AutoSubBusString )  &&     // the first user-assigned net name (if any) will be the name for the complete net
               ( J->GetString() != AutoSubBusString )     )  I->ReplaceDesignator( *J );
          I->AddNet( J.operator->() );
          J = NetList.erase( J );
//          AnyMerged = true;
        }
        else J ++;
      }
    }
//  }

// *** alternatively, the AnyMerged code above can be used and the code below eliminated ***

// Pass 2:  complete pass performs any remaining merges
  // merge nets that share pins
  for ( list<TNet>::iterator I = NetList.begin(); I != NetList.end(); ) {
    bool AddedTwo = false;
    list<TNet>::iterator J = I; ++ J;   // operator+() not defined for list iterators
    while ( J != NetList.end() ) {
      if ( I->CommonPins( J.operator->() ) ) {
        if ( I->DesignatorEquals( NoConnect )  &&  J->GetPartPinSpecCount() > 1 )
          BEGERR << "TNetList::Cleanup() found net " << J.operator->() << " connected to the global no-connect net " << NoConnect << ENDERR;
        if ( ( I->GetString() == AutoSubBusString )  &&     // the first user-assigned net name (if any) will be the name for the complete net
             ( J->GetString() != AutoSubBusString )     )  I->ReplaceDesignator( *J );
        I->AddNet( J.operator->() );
        if ( J->GetPinSpecCount() > 1 )  AddedTwo = true;
        J = NetList.erase( J );
        if ( AddedTwo )  break;
      }
      else J ++;
    }
    if ( ! AddedTwo )  ++ I;    // if added two or more pins from a single net, then try I again
  }



// assign extensions (>0) to duplicate net names
  for ( list<TNet>::iterator I = ++ NetList.begin(); I != NetList.end(); ++ I ) {  // for each net except the first
    list<TNet>::iterator J = I;                 // for all previous nets, in reverse order
    while ( J != NetList.begin() ) {
      -- J;
      if ( I->DesignatorEquals( *J ) ) {
        I->SetExtension( J->GetExtension() + 1 );
        break;
      }
    }
  }

  // copy list back to vector
  Nets.assign( NetList.begin(), NetList.end() );
#endif
}




void TNetList::Dump( ostream& Dest ) {
  for ( unsigned n = 0; n < Nets.size(); ++ n ) {                  // for each net
    Dest << "\nnet " << Nets[ n ] << ":  ";
    for ( unsigned p = 0; p < Nets[ n ].GetPinSpecCount(); ++ p ) {  // for each pin
      if ( ! Nets[ n ].PinSpecs[ p ].OwnerIsPart() )  continue;      // ignore PinSpec's for Module ports
      Dest << ' ' << Nets[ n ].PinSpecs[ p ];
    }
  }
}

int TNetList::DumpSingleNodeNets( ostream& Dest ){
  int NetCount = 0;
  for ( unsigned i = 0; i < GetNetCount(); ++ i ) {
    TNet* N = GetNet( i );
    if ( N->GetPartPinSpecCount() == 1 ) {
      Dest << "\n  net " << *N << ":  ";
      N->DumpAsPortRanges( Dest, "" );
      NetCount ++;
    }
  }
//  Dest << '\n';
  return NetCount;
}

int TNetList::DumpNoConnectNet( ostream& Dest ) {
  int PinSpecCount = 0;
  for ( unsigned n = 0; n < GetNetCount(); ++ n ) {   // for each net
    TNet* N         = GetNet( n );
    if ( N->DesignatorEquals( NoConnect ) ) {
      N->DumpAsPortRanges( Dest );
      PinSpecCount = N->GetPartPinSpecCount();
      break;
    }
  }
  return PinSpecCount;
}


// DumpAllSorted() is mainly intended for regression testing
// The netlist is copied to a temporary netlist which is then sorted
// If DumpAll is true, then PinSpec's of non-Part ports are included in the dump
void TNetList::DumpSorted( ostream& Dest, bool DumpAll ) {
  TNetList NL = *this;     // copy entire netlist to temporary
  NL.SortNets( false );    // sort nets alphabetically by name -- false --> preserve existing net indexes in design's TPort's
  NL.SortPinSpecs();       // sort each net's PinSpecs by port registration order and index
  vector<TNet>& NL_Nets = NL.Nets;
  for ( unsigned n = 0; n <   NL_Nets.size(); ++ n ) {                    // for each net
    TNet& Net = NL_Nets[ n ];
    Dest << "\nnet "  <<      Net << ":  ";
    for ( unsigned p = 0; p < Net.GetPinSpecCount(); ++ p ) {  // for each PinSpec
      TPinSpec& PS = *Net.GetPinSpec( p );
      if ( ! ( DumpAll ||  PS.OwnerIsPart() ) )  continue;
      TPortRange PR;
      PR.SetPort(  PS.GetPort()  );
      PR.SetRange( PS.GetIndex(), PS.GetIndex() );
      Dest << "  " << PR;  // dump the port range
      TPart* Part = dynamic_cast<TPart*>( PS.GetPort()->Name.Owner );
      if ( Part ) {
        Dest << "(";
        Dest << Part->GetReferenceDesignator();  // should always work
        Dest << '.';
        PS.OutputPins( Dest, true );   // true = brief mode
        Dest << ')';
      }
    }
  }
}


void TModPar::AddNonWiredPins( TNet* Dest, TNetList* NL ) {
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    Member->AddNonWiredPins( Dest, NL );
  }
}


void TPort::AddNonWiredPins( TNet* Dest, TNetList* NL ) {
  if ( ! CheckPart() )  return;   // only Part Ports have physical pins
  for ( int i = GetMinRange(); i <= GetMaxRange(); ++ i ) {
    if ( NL->PinInNetList( TPinSpec( this, i ) ) )  continue;
    Dest->AddPin( this, i );
    if ( MultiPinNode() ) return;                           // dump pin once if multi-pin node
  }
}


// Add to Dest any of this module's non-listed PortRanges.
// A listed PortRange is a port or portion of a port that:
//    1. appears in the global SubBusList, OR
//    2. appears in this module's SubBusList AND in this module's owner's SubBusList
// For internal ports (used to implement TModule::bus()), rule 2 is changed to:
//    2. appears in this module's SubBusList

void TModPar::AddNonListedPortRanges( TSubBusList* Dest ) {
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    TModule* M = dynamic_cast<TModule*>( Member );
    if ( M ) {
      M->AddNonListedPortRanges( Dest );  continue;
    }
    TBundle* B = dynamic_cast<TBundle*>( Member );
    if ( B ) {
      B->AddNonListedPortRanges( Dest );  continue;
    }

    TPort* P = dynamic_cast<TPort*>( Member );
    if ( ! P )  continue;                    // should never happen

    TSubBusList* LocalList = dynamic_cast<TSubBusList*>( this );            // zero if TPart (or TBundle, but see below)
    TSubBusList* OwnerList = dynamic_cast<TSubBusList*>( GetOwner() );

    if ( dynamic_cast<TBundle*>( this ) ) {                                 // special case:  TBundle
      TModPar*  UltimateOwner = GetUltimateOwner( this );
      LocalList = dynamic_cast<TSubBusList*>( UltimateOwner             );  // "local list" is actually the  list of this's ultimate owner module
      OwnerList = dynamic_cast<TSubBusList*>( UltimateOwner->GetOwner() );  // "owner list" is actually the owner of this's ultimate owner module
    }

    TSubBusList  EmptyList;
    if ( OwnerList == 0 )  OwnerList = &EmptyList;  // if no owner, use empty list as owner list

  // try full PortRange first
    bool InternalPort = ( P->GetPortType() == ptInternal );
    bool LocalOverlap = true;   // pretend that local list test passed for TPart
    bool OwnerOverlap = true;   // pretend that owner list test passed for internal ports
    if (  LocalList    )  LocalOverlap = ( LocalList->PortOverlap( *P ) != 0 );
    if ( !InternalPort )  OwnerOverlap = ( OwnerList->PortOverlap( *P ) != 0 );

    if (  ( ( ! LocalOverlap       )  ||                     // if no local or no owner overlap
            ( ! OwnerOverlap       )     )  &&
            ( ! GlobalModule()->PortOverlap( *P ) )  ) {        // and no global overlap
      Dest->AddPortRangeSubBus(           *P );              // then add entire range to Dest
      continue;
    }


  // else check each index
    for ( int i = P->GetMinRange(); i <= P->GetMaxRange(); ++ i ) {
      TPortRange PR = (*P)( i );
      if (  LocalList    )  LocalOverlap = ( LocalList->PortOverlap( PR ) != 0 );
      if ( !InternalPort )  OwnerOverlap = ( OwnerList->PortOverlap( PR ) != 0 );
      if ( LocalOverlap  &&  OwnerOverlap    )  continue;  // must be in local SubBusList AND owner's SubBusList, OR
      if ( GlobalModule()->PortOverlap( PR ) )  continue;  //         in global SubBusList
      Dest->AddPortRangeSubBus(         PR );              // else add it to Dest
    }
  }
}


// special-purpose function, not of general use
// see TModule::AddNonListedPortRanges()
// Add a SubBus to the SubBusList, then add aPortRange to the new SubBus
void TSubBusList::AddPortRangeSubBus( TPortRange aPortRange ) {
  SubBuses.push_back( TSubBus() );
  TSubBus* CBB = &SubBuses.back();
  CBB->SetDesignator( "NonListed Port Range" );
  CBB->Add( aPortRange );
}


void TNet::DumpAsPortRanges( ostream& Dest, char* Separator ) {
  TPortRange Temp;
  for ( unsigned i = 0; i < PinSpecs.size(); ++ i ) {
    if ( !         PinSpecs[ i ].OwnerIsPart() )  continue;      // ignore PinSpec's for Module ports
    Temp.SetPort(  PinSpecs[ i ].GetPort()   );
    Temp.SetRange( PinSpecs[ i ].GetIndex(),
                   PinSpecs[ i ].GetIndex()  );
    Dest << Separator << Temp;
  }
}


void TNet::AddNet( TNet* aNet ) {   // add pins from aNet (except duplicates)
  for ( unsigned i = 0; i < aNet->GetPinSpecCount(); ++ i ) {
    TPinSpec* PS = aNet->GetPinSpec( i );
    if ( ! PinInNet( *PS ) )  AddPin( *PS );   // if not already in net, add it
  }
}


TPinSpec*  TNet::CommonPins( TNet* aNet ) {   // return PinSpec of pin common to both this net and aNet
  for ( unsigned i = 0; i < PinSpecs.size(); ++ i ) {           // for all PinSpec's in this net
    for ( unsigned j = 0; j < aNet->GetPinSpecCount(); ++ j ) { // for all PinSpec's in the other net
      TPinSpec* PS = aNet->GetPinSpec( j );
      if ( *PS  ==  PinSpecs[ i ] )  return PS;
    }
  }
  return 0;
}



// true if owner of port is a TPart
bool TPort::CheckPart() { return ( dynamic_cast< TPart* >( GetUltimateOwner( this ) ) != 0 ); }    // >>> should it be legal for TPart's to contain TBundles?


void TPart::AssignReferenceDesignator( TDesignator D ) {
  dynamic_cast<TModule*>( GetOwner() )->AddReferenceDesignator( D );
  ReferenceDesignator = D;
}


// add the reference designator to the appropriate DesignatorList
// this function is typically called only when reference designators are read in a from a file
void TModule::AddReferenceDesignator( TDesignator& D ) {
  if ( ( GetOwner() == 0 )  ||  ( GetReferenceBase() != 0 ) ) {   // if root module or a local reference base has been specified...
    char* dbase = D.GetBase();
    char* fbase = GetFullReferenceBase();
    if ( dbase == 0 )  dbase = "";         // convert null pointers to empty strings for test below
    if ( fbase == 0 )  fbase = "";
    if ( strcmp( dbase, fbase ) != 0 )
      BEGERR << "AddReferenceDesignator(): reference designator base '" << dbase << "' is not equal to full reference base '" << fbase << "'" << ENDERR;

    References.AddDesignator( D );                               // add the reference designator to this module's designator list
    return;
  }

  dynamic_cast<TModule*>( GetOwner() )->AddReferenceDesignator( D );    // else, pass the designator up
}


void  TModule::SetReferenceBase( char* aReferenceBase ) {
  if ( aReferenceBase == 0 )  aReferenceBase = "";  // just so illegal string can be reported
  if ( aReferenceBase[ 0 ] == 0 )
    BEGERR  <<  "SetReferenceBase(): illegal string '"  <<  ReferenceBase  <<  "' for reference base"  <<  ENDERR;
  ReferenceBase = aReferenceBase;
}


void  TModule::CopyReferenceBase( const char* Source ) {     // character array located at Source will be copied to a local persistent character array
  if ( Source == 0 )  Source = "";                           // just so illegal string can be reported
  if ( Source[ 0 ] == 0 )
    BEGERR  <<  "CopyReferenceBase(): illegal string '"  <<  Source  <<  "' for reference base"  <<  ENDERR;
  ReferenceBase = PermString( Source );                      // make a persistent copy of Source and set ReferenceBase to point at it
}



// GetFullReferenceBase() will never return a null pointer
char* TModule::GetFullReferenceBase() {                      // build FullReferenceBase by calling owner's GetFullReferenceBase()
  if ( FullReferenceBase != 0 )  return FullReferenceBase;   // this function was already called

  TModule* MyOwner = dynamic_cast<TModule*>( GetOwner() );   // else build FullReferenceBase using Owner's FullReferenceBase and this module's ReferenceBase
  char* LocalBase  = "";
  char* OwnerBase  = "";                                     // in case this module is root module (MyOwner is 0)
  if ( ReferenceBase )  LocalBase = ReferenceBase;
  if ( MyOwner )        OwnerBase = MyOwner->GetFullReferenceBase();

  string Hier       = OwnerBase;
  string Local      = LocalBase;
  string Full       = Hier + Local;
  FullReferenceBase = PermString( Full.c_str() );            // pointer to permanent null-terminated char array
  return FullReferenceBase;
}

void TModule::AssignReferenceDesignators() {

// only parts get reference designators, not modules
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    TModPar* M = dynamic_cast< TModPar* >( Member );      // member Modules and Parts
    if ( M ) M->AssignReferenceDesignators();             // descend into Module or assign reference to Part
  }
}

void TPart::AssignReferenceDesignators() {
  if ( ReferencePrefix == 0 )
    BEGERR << "Part " << *this << " has no reference designator prefix.\nEach part class must assign a reference designator prefix,\ne.g., \"R\" for resistors, \"U\" for IC's, etc.,\nvia TPart::SetReferencePrefix()" << ENDERR;
  if ( ! ReferenceDesignator.Assigned() ) {     // reference designator may already be assigned, e.g., by ReadReferenceDesignators()
    ReferenceDesignator = GetOwner()->SupplyReferenceDesignator( ReferencePrefix );
  }
}

 // supply caller with a unique reference designator
TDesignator TModule::SupplyReferenceDesignator( char* aReferencePrefix ) {
  if ( ( GetOwner() == 0 )  ||  ( GetReferenceBase() != 0 ) )                              // if root module or a local reference base has been specified...
    return References.GetUniqueDesignator( GetFullReferenceBase(), aReferencePrefix, 1 );  // generate a unique reference designator
  else
    return GetOwner()->SupplyReferenceDesignator( aReferencePrefix );  // get reference designator from owner
}

TDesignator TPart::SupplyReferenceDesignator( char* aReferencePrefix ) {
  return  "ERR";   // should never happen
}

// reference designators can be written to a file for future use by AssignReferenceDesignators()
//   file format:   symbolic_name       base  prefix  number
//   e.g.:          PowerModule.BigCap  PW  C  1

void TModule::WriteReferenceDesignators( ofstream& Dest ) {
  for ( TCBBase* CBB = IterFirst(); CBB != 0; CBB = IterNext() ) {
    TPart* Part = dynamic_cast<TPart*>( CBB );
    if ( Part ) {
      TDesignator D = Part->GetReferenceDesignator();

      if ( ! D.Assigned() )
        BEGERR << "WriteReferenceDesignators() encountered non-assigned designator for part " << *Part << ENDERR;  // should never happen

// check that reference prefixes and bases are legal
// here may be the best place to do this, because the offending part can be identified by full name
      char* Prefix    = D.GetString();   // reference prefix
      char* Base      = D.GetBase();     // reference base
      char* ErrString = "prefix";        // used in error reporting
      try {
         IsValidNameFragment( Prefix, false );  // check, do not ignore empty
         ErrString = "base";
         if ( Base == 0 )  Base = "";           // convert null string to empty string
         IsValidNameFragment( Base,   true  );  // check, ignore empty
      }
      catch ( TE_NameFragmentNull     ) { BEGERR << "WriteReferenceDesignator() encountered non-defined reference "                             << ErrString << " for part "  << *Part << ENDERR; }
      catch ( TE_NameFragmentChar   E ) { BEGERR << "WriteReferenceDesignator() encountered illegal character "     << E.C << " in reference "  << ErrString << " for part "  << *Part << ENDERR; }
      catch ( TE_NameFragmentLength E ) { BEGERR << "WriteReferenceDesignator() encountered illegal length of "     << E.L << " for reference " << ErrString << " for part "  << *Part << ENDERR; }

      Dest                    <<  *Part          <<  "  ";   // full symbolic name
      if ( *Base == 0 )  Dest <<  NoBaseString   <<   " ";   // base
      else               Dest <<  Base           <<   " ";
      Dest                    <<  Prefix         <<   " ";   // prefix
      Dest                    <<  D.GetNumber()  <<  "\n";   // number
    }
  }
};

// read in previous reference designators to preserve reference designators of a previous design iteration
// a part's full symbolic name (i.e., path) associates the part with the reference designator in the file, e.g.,
//   PowerModule.BigCap    PW  C  1       <-- example line in Srce
//   ------------------   -----------
//   full symbolic name   ref. desig.
// symbolic names in Srce that no longer appear in the current design are ignored
struct TReadHelper {
  TPart*   P;   // pointer to part
  string   N;   // symbolic name of part
};

int TModule::ReadReferenceDesignators( ifstream& Srce ) {
  if ( Srce.fail() )  return 0;   // >>> assume there is no file to read in

  vector<TReadHelper> Parts;    // read parts into a vector of TReadHelper's
  for ( TCBBase* CBB = IterFirst(); CBB != 0; CBB = IterNext() ) {
    TPart* Part = dynamic_cast<TPart*>( CBB );
    if ( Part ) {
      TReadHelper    Temp;
      ostringstream  PartName;  PartName << *Part;      // get part's full name
      Temp.N = PartName.str();
      Temp.P = Part;
      Parts.push_back( Temp );
    }
  }

  int Line = 0;
  int SuccessCount = 0;
  while ( ! Srce.eof() ) {
    Line ++;
    string  Name;
    string  Base;
    string  Prefix;
    int     Number;
    Srce >> Name;      // read in whitespace-terminated word
    Srce >> Base;
    Srce >> Prefix;
    Srce >> Number;
    if ( Srce.fail() )
      BEGERR << "Could not read reference designator file, line " << Line << ENDERR;

    for ( unsigned i = 0; i < Parts.size(); ++ i ) {
      if ( Parts[ i ].N == Name ) {                                          // if part's name matches name read from file...
        char*                        prefix = PermString( Prefix.c_str() );  // create permanent copies
        char* base = 0;                                                      // null base is legal
        if ( Base != NoBaseString )  base   = PermString( Base.c_str()   );  // if base is equal to the string used to indicate no base, e.g., "<none>"
        TDesignator D( prefix, Number, 0, base );                            // assign it the designator from the file
        Parts[ i ].P->AssignReferenceDesignator( D );
        SuccessCount ++;
        break;
      }
    }
    ws( Srce );       // skip whitespace
  }
  return SuccessCount;
}


TDesignator TDesignatorList::GetUniqueDesignator( char* aDesignatorBase, char* aDesignatorPrefix, int aStartNumber ) {
  Sort();                    // algorithm below depends on a sorted list
  TDesignator Temp( aDesignatorPrefix, aStartNumber, 0, aDesignatorBase );

  bool  FoundString = false;

  for ( vector<TDesignator>::iterator I = Designators.begin(); I != Designators.end(); I++ ) {
    if ( I->BaseEquals( Temp )  &&  I->StringEquals( Temp ) ) {
      FoundString = true;
      if ( Temp.GetNumber()  <  I->GetNumber() ) {
        Designators.insert( I, Temp );               // non-typical case: utilize a number that has not yet been used with this String and Base
        return Temp;
      }
      Temp.ReplaceNumber( I->GetNumber() + 1 );      // typical case:  try a number one higher, typically until number exceeds all other numbers for this Base and String
      continue;
    }
    if ( FoundString ) {   // if string was previously found, but now a different string is encountered
      Designators.insert( I, Temp );  // typical case:  add after all other designators having same Base and String
      return Temp;
    }
  }
  AddDesignator( Temp );     // Temp was the first designator in the list having its Base and String
  return Temp;
}

void  TModule::FillDesignatorList( TDesignatorList* DesignatorList ) {
  for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
    TModPar* MP = dynamic_cast< TModPar* >( Member );
    if ( MP ) MP->FillDesignatorList( DesignatorList );
  }
}

void  TPart::FillDesignatorList( TDesignatorList* DesignatorList ) {
  DesignatorList->AddDesignator( ReferenceDesignator );
}

void TModule::CheckReferenceDesignators() {
  TDesignatorList D;
  FillDesignatorList( & D );

  vector<string>      StringList;
  D.AddToStringList( &StringList );               // create string version of D
  sort( StringList.begin(), StringList.end() );   // sort so that comparison is trivial

  for ( unsigned i = 0; i < StringList.size() - 1; ++ i ) {  // compare each designator with the next one in the list
    if ( StringList[ i ]  ==  StringList[ i + 1 ] )
      BEGERR << "CheckReferenceDesignators() found duplicate designator: " << StringList[ i ] << ENDERR;
  }
}



char* DateTimeString() {
static char DateTime[ 32 ];
   time_t t;
   struct tm *area;

   t = time(NULL);
   area = localtime(&t);
   strncpy( DateTime, asctime(area), 30 );
   DateTime[ 24 ] = 0;   // get rid of newline
   return DateTime;
}


struct TBOM_Helper {
  const char* TypeName;
  TPackage    Package;
  int         Count;
  TBOM_Helper( const char* aTypeName = "", TPackage aPackage = TPackage() , int aCount = 0 ) {
    TypeName = aTypeName; Package = aPackage; Count = aCount;
  }
};
bool operator<( TBOM_Helper A, TBOM_Helper B ) { return strcmp( A.TypeName, B.TypeName ) < 0; }

// output bill of materials
// Part Type, Package, Qty, Ref Des

void TModule::OutputBOM( ofstream& Dest ) {

  vector<TBOM_Helper> Helpers;
  // get list of all parts in the design
  vector<TPart*> Parts;
  for ( TCBBase* CBB = IterFirst(); CBB != 0; CBB = IterNext() ) {
    TPart* Part = dynamic_cast<TPart*>( CBB );
    if ( Part ) Parts.push_back( Part );
  }
  for ( unsigned p = 0; p < Parts.size(); ++ p ) {
    const char* TypeName = Parts[ p ]->GetTypeName();   // type of part
    unsigned h;
    for ( h = 0; h < Helpers.size(); ++ h ) {
      if ( strcmp( TypeName, Helpers[ h ].TypeName ) == 0 ) {
        Helpers[ h ].Count ++;
        break;
      }
    }
    if ( h < Helpers.size() )  continue;
    Helpers.push_back( TBOM_Helper( TypeName, Parts[ p ]->GetPackage(), 1 ) );
  }

  stable_sort( Helpers.begin(), Helpers.end() );   // sort alphabetically by PartType
  Dest << "Bill of Materials\n";
  Dest << "Part Type, Package, Quantity,Reference Designators\n";
  for ( unsigned h = 0; h < Helpers.size(); ++ h ) {
    Dest << Helpers[ h ].TypeName << ", "
         << Helpers[ h ].Package  << ", "
         << Helpers[ h ].Count;
    vector<TDesignator> Designators;

    Designators.clear();
    for ( unsigned p = 0; p < Parts.size(); ++ p ) {
      if ( strcmp( Parts[ p ]->GetTypeName(), Helpers[ h ].TypeName ) == 0 ) {
        Designators.push_back( Parts[ p ]->GetReferenceDesignator() );
      }
    }
    sort( Designators.begin(), Designators.end() );

    Dest << ",\"";        // comma and opening quote (no intervening space->designators loaded into single column of spreadsheet)
    bool First   = true;  // do not output comma for first reference designator
    TDesignator Working;  // used to detect consecutive designators
    int Consecutive = 0;
    for ( unsigned d = 0; d < Designators.size(); ++ d ) {
      TDesignator D = Designators[ d ];
      if ( Consecutive == 0 ) {
        if ( ! First ) Dest << ", ";
        Dest << D;
        Working = D;
        Working ++;        // next consecutive designator
        Consecutive ++;
        First = false;
        continue;
      }
      if ( D.DesignatorEquals( Working ) ) {
        Working ++;
        Consecutive ++;
        continue;
      }
      if ( Consecutive > 1 ) {
        Working --;      // go back to last consecutive designator
        if ( Consecutive > 2 )  Dest << "-";
        else                    Dest << ", ";
        Dest << Working;
      }
      Consecutive = 0;
      d --;              // do same designator again
    }

    if ( Consecutive > 1 ) {
      Working --;      // go back to last consecutive designator
      if ( Consecutive > 2 )  Dest << "-";
      else                    Dest << ", ";
      Dest << Working;
    }

    Dest << "\"\n";   // closing quote and newline
  }
}

