/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_base.h:  CBOLD fundamental classes

#ifndef _cb_baseH_
#define _cb_baseH_

//#define DEBUG_ON   // for progress messages

#include "time.h"       // for time(), used by TLogStatus
#include <ctype.h>      // for is...()
#include <typeinfo>     // for RTTI
#include <algorithm>    // for sort
#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <vector>
#include <list>

using std::string;
using std::vector;
using std::sort;
using std::stable_sort;
using std::ostream;
using std::ofstream;
using std::ifstream;
using std::ostringstream;
using std::cout;
using std::cin;
using std::ios_base;
using std::flush;
using std::noskipws;

// the names below are #define'd in header files supplied with some compilers -- as they are common port names for TPart's, they are undefined below
#undef OUT
#undef IN

#include "cb_user_pref.h"   // user preferences
#include "cb_char_bucket.h"
#include "cb_imp_spec.h"    // compiler-specific code

// convenient way to create identical parts with different names,
// e.g., CP_R10K, CP_R100, CP_R1M
#define clone( aProto, aNew ) class aNew : public aProto {}

// macros that register members of a module:
//   reg   -- register a single member with a simple designator (no Number)
//   regn  -- register a single element of an array member with a full designator (string + Number) (rarely used)
//   rega  -- register an array member
//   regb  -- register a bus port (with left and right range) with a simple designator
//   regab -- register an array of bus ports (with left and right range)

#define reg(   aName )          Reg( &aName, #aName )               // #aName stringifies aName, i.e., wraps it in quotes
#define regn(  aName, aNumber ) Reg( &aName[ aNumber ], #aName, aNumber )
#define rega(  aName, aCount ) { for ( int i = 0; i < aCount; ++i ) regn( aName, i ); }
#define regb(  aName, aL, aR ) { reg( aName ); aName.SetRange( aL, aR ); }
#define regab( aName, aCount, aL, aR ) { for ( int i = 0; i < aCount; ++i ) { regn( aName, i ); aName[ i ].SetRange( aL, aR ); } }


char* PermString( const char* CP );   // return a pointer to a permanent (non-local) copy of string pointed to by CP

// CSTRING() is a utility for forming a char* using stream-like syntax.
// The char* points to a null terminated character array and is valid forever, i.e., it behaves just like a char* created using quoted text.
// In the example below, if i has values 1, 2, 3, MyPointer will point to "Foo1Bar", "Foo2Bar", "Foo3Bar"
//   char* MyPointer = CSTRING( "Foo" << i << "Bar" )

char* string_to_cstr( ostream& OS );
//#define CSTRING( A ) string_to_cstr( ostringstream()         << A )  // don't know why this doesn't work--seems that char* in A is printed in hex rather than as text
  #define CSTRING( A ) string_to_cstr( ostringstream().flush() << A )


// SCSTRING() is the same as CSTRING() except one static string is used as storage for all calls, i.e., the pointer returned is only valid until the next call of SCSTRING().
const char* string_to_scstr( ostream& OS );
  #define SCSTRING( A ) string_to_scstr( ostringstream().flush() << A )



// these call ...fstream::open() -- if the file cannot be opened, they use BEGERR/ENDERR to output an error message via an exception
void OpenFileI(  ifstream& Stream, const char* Path );
void OpenFileO(  ofstream& Stream, const char* Path );

extern char* AutoSubBusString;                                // string used for auto-named subbuses
// types of ports
// the default is ptUnspecified
// ptInternal is used when TModule::bus() creates a SubBus based only on a name and range
enum    TPortType  { ptIn, ptOut, ptInOut, ptUnspecified, ptInternal };


//class        ECBBase;          // base of all exceptions
class          TCBBase;
class          TModPar;
class          TDesignator;
class          TPackage;
class          TPortRange;
class          TRange;
class          TModule;
class          TRootModule;
class          TPart;
class          TBundle;
class          TPort;
class          TInPort;
class          TOutPort;
class          TInOutPort;
class          TNet;
class          TNetList;
class          TPinSpec;
class          TNetName;
class          TSubBusList;
typedef char*  TPin;          // char* may be ok for TPin;

typedef TPort          port;
typedef TInPort       iport;
typedef TOutPort      oport;
typedef TInOutPort   ioport;


typedef TPart        part;
typedef TModule      module;

ostream& operator<<( ostream& Dest, TPackage     P   );
ostream& operator<<( ostream& Dest, TDesignator  D   );
ostream& operator<<( ostream& Dest, TPortRange   PR  );
ostream& operator<<( ostream& Dest, TPinSpec     PS  );
ostream& operator<<( ostream& Dest, TCBBase      CBB );
ostream& operator<<( ostream& Dest, TRange*      R   );
ostream& operator<<( ostream& Dest, TSubBusList* SBL );
ostream& operator<<( ostream& Dest, TNetName*    NN  );
ostream& operator<<( ostream& Dest, TNetName     NN  );
ostream& Out_PR_NP(  ostream& Dest, TPortRange   PR  );  // output port range without full path
ostream& Out_D_NP(   ostream& Dest, TDesignator  D   );  // output designator without full path
ostream& operator<<( ostream& Dest, TModule*     M   );

void MessageIfError( ostream& Dest, TCBBase* ErrorSource, char* PassName );   // used only in main()'s exception handler


bool operator<( TDesignator A, TDesignator B );  // use with stdlib's sort()
bool operator<( TNetName    A, TNetName    B );  // use with stdlib's sort()


// Reg() for highest-level module
void  Reg( TRootModule* RootModule, char* aName, int aNumber = -1 );

// handy utilities
inline TModPar* GetUltimateOwner( TCBBase* CBB );   // inline of this function increases overall program speed considerably

class TE_NameFragmentNull {};
class TE_NameFragmentEmpty {};
class TE_NameFragmentChar {
public:
  char C;  // the offending character
  TE_NameFragmentChar( char aC ) { C = aC; }
};
class TE_NameFragmentLength {
public:
  int L;   // the offending length
  TE_NameFragmentLength( int aL ) { L = aL; }
};
void IsValidNameFragment( char* S, bool IgnoreEmpty = false ) throw ( TE_NameFragmentNull, TE_NameFragmentEmpty, TE_NameFragmentChar, TE_NameFragmentLength );


// macros for conveniently throwing an error message
// example:  if ( error )  BEGERR << "There was an error" << ENDERR;
//#define  BEGERR  { ECBBase TempE_; TempE_
//#define  ENDERR   '.'; throw ECBBase.GetMessage; }

#define  BEGERR  { ostringstream Temp_OStringStream_; Temp_OStringStream_
#define  ENDERR   '.'; throw Temp_OStringStream_.str(); }


// !!! possibly attempt to identify the offending TModPar
void CBOLD_CopyError( char* ClassName, char* Mode);


// complex designator example:
//   VMU5  consists of three parts:  VM      U        5
//                                  ----   ------   ------
//                                  Base   Prefix   Number
//                                           or
//                                         String

class TDesignator {     // e.g., MyModule1
  char*    Base;        // typically only used with complex designators
  char*    String;      // e.g., MyModule
  int      Number;      // e.g., 1          Number of -1 indicates no Number
public:
  TModPar*   Owner;         // owner if designator applies to a member of a module -- needed only for full path output
  bool     Assigned()   { return Registered();   }  // alias for Registered()
  bool     Registered() { return( String != 0 ); }
  char*    GetBase()    { return Base; }
  char*    GetString()  { if ( ! Registered() ) return "NO_NAME_REGISTERED";  return String; }
  int      GetNumber()  { return Number; }
//  TDesignator() { String = 0; Number = -1; Owner = 0; }
  TDesignator( char* aString = 0, int aNumber = -1, TModPar* aOwner = 0, char* aBase = 0 ) {
    String = aString; Number = aNumber; Owner = aOwner; Base = aBase;
  }

  TDesignator& operator++()      { ++ Number; return *this; }                           // prefix
  TDesignator  operator++( int ) { TDesignator Temp = *this; ++ Number; return Temp; }  // postfix
  TDesignator& operator--()      { -- Number; return *this; }                           // prefix
  TDesignator  operator--( int ) { TDesignator Temp = *this; -- Number; return Temp; }  // postfix

  bool DesignatorEquals( const char* aString, int aNumber = -1 ) {
    if ( Number != aNumber              ) return false;
    if ( strcmp( String, aString ) != 0 ) return false;
    return true;
  }
  bool DesignatorEquals( TDesignator& aDesignator ) {
    if ( Number != aDesignator.Number              ) return false;
    if ( strcmp( String, aDesignator.String ) != 0 ) return false;
    return BaseEquals( aDesignator );
  }
  bool BaseEquals( TDesignator& aDesignator ) {
    if ( Base == aDesignator.Base )                  return true;   // e.g., if both are null
    if ( Base == 0  ||  aDesignator.Base == 0 )      return false;
    return  strcmp( Base, aDesignator.Base ) == 0;
  }
  bool BaseEquals( const char* aBase ) {
    if ( Base == aBase )                     return true;   // e.g., if both are null
    if ( Base == 0  ||  aBase == 0 )         return false;
    return  strcmp( Base, aBase ) == 0;
  }
  bool StringEquals( const char* aString ) {
    if ( strcmp( String, aString ) != 0 ) return false;
    return true;
  }
  bool StringEquals( TDesignator& aDesignator ) {
    return  StringEquals( aDesignator.String );
  }
  void SetDesignator( char* aString, int aNumber = -1, TModPar* aOwner = 0, char* aBase = 0 )  {
    if ( String )
      BEGERR << "Attempt to rename object " << *this << ENDERR;
    String = aString; Number = aNumber; Owner = aOwner; Base = aBase;
  }
  void SetDesignator( TDesignator& aDesignator ) {
    SetDesignator( aDesignator.String, aDesignator.Number, aDesignator.Owner, aDesignator.Base );
  }
  void ReplaceNumber( int aNumber = -1 ) {   // only way for user to replace Number
    Number = aNumber;
  };
  void ReplaceString( char* aString ) {      // only way for user to replace String
    String = aString;
  };
  void ReplaceBase( char* aBase ) {
    Base = aBase;
  }
  void ReplaceDesignator( TDesignator& D ) {  // same as SetDesignator, but no exception
    String = D.String; Number = D.Number; Owner = D.Owner; Base = D.Base;
  }
  string AsString() {  // return designator as a single standard string
    string Temp;
    if ( Base )   Temp += string( Base );
    if ( String ) Temp += string( String );
    int  N = GetNumber();
    if ( N != -1 ) {
      Temp += SCSTRING( N );    // convert number to string
    }
    return Temp;
  }
};


class TDesignatorList {
  bool Sorted;
  vector<TDesignator> Designators;   // list of designators, possibly in random order
public:
  TDesignatorList() { Sorted = false; }
  void Sort() { if ( ! Sorted )  sort( Designators.begin(), Designators.end() ); }
  TDesignator GetUniqueDesignator( char* aDesignatorBase, char* aDesignatorPrefix, int aStartNumber = 0 );
  void AddDesignator( TDesignator& aDesignator ) {
    Designators.push_back( aDesignator );
    Sorted = false;
  }
  void AddToStringList( vector<string>* aStringList ) {     // add designators as strings to aStringList
    for ( unsigned i = 0; i < Designators.size(); ++ i ) {
      aStringList->push_back( Designators[ i ].AsString() );
    }
  }

};


class TCBBase {       // base for TModPar and TPort -- only TCBBase's may be registered members of TModPar's
  char*  TypeName;   // user may override type name, e.g., to change a part value from CP_RCHIP805 to CP_R30K
public:
  TDesignator Name;
  int         SerialNumber;         // registration serial number -- can be used to sort TCBBase's in the order they were registered
  TCBBase*     Prev;                 // linked list of members of Name.Owner (i.e., this TCBBase's owner)
  TCBBase*     Next;
  TCBBase() {  Name.Owner = 0; Prev = 0; Next = 0; TypeName = 0; }
  TModPar* GetOwner() { return Name.Owner; }   // return the immediate owner

  void     ErrIfRegistered( char* Message ) {
    if ( Name.Registered() )  BEGERR << Message << ": not allowed after registration" << ENDERR;
  }

  virtual void SetTypeName( char* aTypeName ) {
    if ( TypeName != 0 )
      BEGERR << "Attempt to rename " << *this << " to " << aTypeName << ENDERR;
    TypeName = aTypeName;
  }

  virtual const char* GetTypeName() {     // user may override GetTypeName() -- RTTI class names are implementation-specific
    if ( TypeName ) return TypeName;
    return Translate_RTTI_Name( typeid( *this ).name() );  // name() returns char array with compiler-defined contents -- see imp_spec.h/cpp
  }
  
  virtual void DumpAll( ostream& Dest ) { Dest << "\nTCBBase" << Name; }
  virtual void AddNonWiredPins( TNet* Dest, TNetList* NL ) {}           // add pins that do not appear in the netlist to Dest net
};


class TRange {
  virtual void Dummy() {}                      // to make descendents polymorphic
protected:
  int           Count;         // number of individual wires
  int           LeftRange;     // left index range for bus, else 0
  int           RightRange;    // right index range for bus, else 0
  int           MinRange;      // minimum of LeftRange and RightRange
  int           MaxRange;      // maximum of LeftRange and RightRange

public:
  bool InRange( int i ) { return ( ( i >= MinRange )  &&  ( i <= MaxRange ) ); }
  bool Equal( TRange& OtherRange ) {
    return ( ( MinRange == OtherRange.MinRange ) &&
             ( MaxRange == OtherRange.MaxRange )     );
  }
  bool Overlap( TRange& OtherRange ) {
    return ( InRange( OtherRange.MinRange ) ||
             InRange( OtherRange.MaxRange ) ||
             ( ( OtherRange.MinRange < MinRange ) &&
               ( OtherRange.MaxRange > MaxRange )     ) );
  }
  int  GetCount()       { return Count; }
  int  GetLeftRange()   { return LeftRange; }
  int  GetRightRange( ) { return RightRange; }
  int  GetMinRange()    { return MinRange; }
  int  GetMaxRange()    { return MaxRange; }
  int  TranslateIndex( const TRange& OtherRange, const int Index ) {   // translate index of one range into index of this range
    if ( OtherRange.Count != Count ) {
      TRange Temp = OtherRange;
      BEGERR << "TranslateIndex() requires same count for two ranges: " << this << " and " << &Temp << ENDERR;
    }
    int Offset = Index - OtherRange.LeftRange;
    if ( Offset < 0 ) Offset = -Offset;           // offset from left (always positive)
    if ( LeftRange == MinRange )  return LeftRange + Offset;
    else                          return LeftRange - Offset;
  }
  void SetRange( int aLeftRange, int aRightRange ) {
    LeftRange  = aLeftRange;
    RightRange = aRightRange;
    if ( LeftRange > RightRange ) { MaxRange =  LeftRange;  MinRange = RightRange; }
    else                          { MaxRange = RightRange;  MinRange =  LeftRange; }
    Count = MaxRange - MinRange + 1;
  }
  TRange() {
    Count  =  1;        // default: single--call SetRange to change to multi
    LeftRange  =  0;
    RightRange =  0;
    MinRange   =  0;
    MaxRange   =  0;
  }
};


class TPortRange : public TRange {
  TPort*  Port;
public:
  bool Defined() const { return Port != 0; }                   // used primarily for default argument passing to TSubBusList::conn()
  bool Registered() const;
  bool Equal( TPortRange OtherPortRange ) {   // true if ports match and ranges match
    if ( OtherPortRange.GetPort() != GetPort() )  return false;
    return TRange::Equal( OtherPortRange );
  }
  bool Overlap( TPortRange OtherPortRange ) {   // true if ports match and ranges overlap
    if ( OtherPortRange.GetPort() != GetPort() )  return false;
    return TRange::Overlap( OtherPortRange );
  }
  TPortRange operator()( int aLeft ) { return operator()( aLeft, aLeft ); }  // single wire shorthand, e.g., P(4)
  TPortRange operator()( int aLeft, int aRight ) {                           // range shorthand, e.g., P(31,0)
    TPortRange Temp;
    Temp.SetRange( aLeft, aRight );
    if ( ( Temp.GetMinRange() < GetMinRange() ) ||
         ( Temp.GetMaxRange() > GetMaxRange() )     )
         BEGERR << "Bad range requested for port " << this << ". Requested range (" << aLeft << "," << aRight << "). Port's range is (" << GetLeftRange() << "," << GetRightRange() << ")" << ENDERR;
    Temp.SetPort( Port );
    return Temp;
  }

  void SetPort( TPort* aPort ) { Port = aPort; }
  TPort* GetPort() { return Port; }
  TPortRange() { Port = 0; }
};


// One or more wires form an SubBus.
// A single TSubBus might not contain all connection information for all wires in the SubBus.
// I.e., a complete netlist can only be formed by examining all SubBuses.
// A SubBus's range is equal to the range of the first connection added to it.
class TSubBus : public TDesignator {
  vector<TPortRange>  Connections;
public:
  TPortRange GetConnection( unsigned c ) {
    if ( c >= Connections.size() )
      BEGERR << "GetConnection(): Bad connection index of " << c << " for SubBus " << *this << ENDERR;
    return Connections[ c ];
  }
  size_t GetConnectionCount() { return Connections.size(); }
  int GetWireCount() {
    if ( Connections.size() == 0 ) return 0;
    return Connections[ 0 ].GetCount();
  }
  TRange GetRange() {
    if ( Connections.size() == 0 )
      BEGERR << "GetRange(): No connections for SubBus " << *this << ENDERR;
    return Connections[ 0 ];
  }
  int GetLeftRange() {
    if ( Connections.size() == 0 )
      BEGERR << "GetLeftRange(): No connections for SubBus " << *this << ENDERR;
    return Connections[ 0 ].GetLeftRange();
  }
  int GetRightRange() {
    if ( Connections.size() == 0 )
      BEGERR << "GetRightRange(): No connections for SubBus " << *this << ENDERR;
    return Connections[ 0 ].GetRightRange();
  }
  int  FindPin(         TPinSpec aPinSpec     );
  TPortRange* PortOverlap( TPortRange aPortRange );
  TPortRange* MemberPortOverlap( TPortRange aPortRange, TModPar* ModPar );


  bool IsConnection( TPortRange aPortRange ) {  // return true if aPortRange is one of the connections in this SubBus
    for ( unsigned i = 0; i < Connections.size(); ++ i ) {
      if ( Connections[ i ].Equal( aPortRange ) )  return true;
    }
    return false;
  }
  void Add( TPortRange aPortRange ) {
// New model for no-connect:  /NC is a single-wire net. User must add each wire individually ( merge() automates this ).
//    if ( ! DesignatorEquals( NoConnect ) ) {   // PortRanges added to no-connect net can be any size
      if ( Connections.size() != 0 ) {
        if ( aPortRange.GetCount() != Connections[ 0 ].GetCount() )
          BEGERR << "TSubBus::Add(): port range mismatch when adding " << aPortRange << " to SubBus " << *this << ENDERR;
      }
//    }
    if ( IsConnection( aPortRange ) )  return;  // *** for efficiency of netlist algorithsm: do not add duplicate PortRange ***
    Connections.push_back( aPortRange );
  }
  void DumpAll( ostream& Dest ) {
    Dest << "\nSubBus " << *this << ": ";
    for ( unsigned i = 0; i < Connections.size(); ++ i )
      Dest << ' ' << Connections[ i ];
  }
  TSubBus() : Connections( 0 ) {}
};



class TSubBusList {                     // list of SubBuses
  virtual void Dummy() {}     // to make descendants polymorphic
protected:
         vector<TSubBus>    SubBuses;
         int                AutoSubBusNumber;  // index used for automatic SubBus naming
public:
  size_t GetSubBusCount() { return SubBuses.size(); }
  TSubBus&  GetSubBus( unsigned i ) {
    if ( i >= GetSubBusCount() )
      BEGERR << "GetSubBus(): bad SubBus index of " << i << " for " << this << ENDERR;
    return( SubBuses[ i ] );
  }
  TPortRange* PortOverlap( TPortRange aPortRange );
  TPortRange* MemberPortOverlap( TPortRange aPortRange, TModPar* ModPar );

  TSubBusList() : SubBuses( 0 ) { AutoSubBusNumber = 1; }   // set initial value for auto SubBus naming
  void           DumpAll( ostream& Dest ){                  // output connection list to Dest
    Dest << "\nSubBusList";
    for ( unsigned i = 0; i < SubBuses.size(); ++ i )
      SubBuses[ i ].DumpAll( Dest );
  }
  void AddPortRangeSubBus( TPortRange aPortRange );   // special-purpose function, not of general use--see TModule::AddNonListedPortRanges()
};


class TSubBusSpec {     // rarely used class -- see sprop.h
public:
  char*    Name;        // name of SubBus
  TModule* Module;      // module whose SubBusList contains the SubBus.
  TSubBusSpec( char* aName, TModule* aModule ) { Name = aName; Module = aModule; }
};


class TModPar : public TCBBase {       // common base for TModule and TPart
protected:
  int       MemberCount;            // count of registered members
  int       PortCount;              // count of registered members that are ports
  TCBBase*  FirstMember;
  TCBBase*  LastMember;
  TCBBase*  IterMember;             // for iterating through hierarchy

public:
  static TModPar*     RegisteringModPar;     // ModPar whose Register() is currently being called
  static TModPar*     PostConnectingModPar;  // ModPar whose PostConnect() is currently being called
  TRootModule*        Root;                  // root module of hierarchy that contains this TModPar
  TModule*            GlobalModule();        // returns the TGlobalSubBusModule (in Root) that holds the subbus list for global nets

  TCBBase*  GetFirstMember() { return FirstMember; }
  TCBBase*  GetLastMember()  { return LastMember; }
  int       GetPortCount()   { return PortCount; }
  TPort*    GetPort( int N );   // get the Nth port ( N = 0 for the first port registered, etc.)

  bool  IsMember( TCBBase* CBB ) {
    for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
      if ( CBB == Member ) return true;
    }
    return false;
  }
  virtual void        FillNetList( TNetList* NetList ) {}
  virtual void        FillDesignatorList( TDesignatorList* DesignatorList ) {}
  virtual TDesignator SupplyReferenceDesignator( char* aReferencePrefix ) = 0;    // supply caller with a unique reference designator
  virtual void        AssignReferenceDesignators() = 0;                           // assign reference designators to members
  virtual void        Register() = 0;
  virtual void        PostConnect() {};
  void                PostConnectAll() {   // call local PostConnect() and then PostConnect() of each member modpar
    PostConnectingModPar = this;      // for error reporting
    PostConnect();                    // local PostConnect()
    PostConnectingModPar = 0;

    for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
      TModPar* MP = dynamic_cast<TModPar*>( Member );
      if ( MP )  MP->PostConnectAll();                // PostConnect() in member
    }
  }


  TModPar() { MemberCount = 0; PortCount = 0; FirstMember = 0; LastMember = 0; }

  void  Reg( TCBBase* aMember, char* aName, int aNumber = -1 );     // register a member with the module
  virtual void DumpAll( ostream& Dest ) = 0;
  virtual void AddNonWiredPins( TNet* Dest, TNetList* NL );         // add non-wired pins to Dest net
  void         AddNonListedPortRanges( TSubBusList* Dest );

  TPort* FindPort( char* aString, int aNumber = -1 );               // find port with specified designator

// iterators for navigating the hierarchy
  TCBBase* IterFirst() {
    for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
      TModPar* TMP = dynamic_cast< TModPar* >( Member );
      if ( TMP ) TMP->IterFirst();
    }
    return IterMember = (TCBBase*)( this );
  }

  TCBBase* IterNext() {
    if ( IterMember == 0                 )  return 0;                          // if already beyond last
    if ( IterMember == (TCBBase*)( this ) )  return IterMember = FirstMember;  // if first call of IterNext() for this module
    TModPar* MP = dynamic_cast< TModPar* >( IterMember );
    if ( MP ) {                              // if current member is a ModPar
      TCBBase* CBB = MP->IterNext();         // use member's IterNext() until it returns zero
      if ( CBB )  return CBB;
    }
    return  IterMember = IterMember->Next;   // go on to next member -- zero if no more members
  }

};




class TModule : public TModPar, public TSubBusList {
  static TPortRange   DefaultRange;        // used as default argument to conn()
  TDesignatorList     References;
                                           // reference base:  first characters of the reference designators of all parts in this module
  char*               ReferenceBase;       // local version, specified by user
  char*               FullReferenceBase;   // full hierarchical version, created by AssignReferenceDesignators()
  int                 wire_int( TBundle& Bundle, TModPar& ModPar );  // called by corresponding wire()
  int                 wire_int( TPort*   Port,   TModPar& ModPar );  // called by above wire_int()

  vector<TPort*>      InternalPorts;       // used only to delete internal ports
  TPort*              NewInternalPort( char* Name, int LeftRange, int RightRange );  // used by TModule::bus()
public:
  static TModule*     ConnectingModule;    // module whose Connect()  is currently being called--see shorthand operators

  TModule()  { ReferenceBase = 0; FullReferenceBase = 0; }
  ~TModule();
  void                SetReferenceBase(  char* aReferenceBase );  // aReferenceBase points persistent character array
  void                CopyReferenceBase( const char* Source );    // character array located at Source will be copied to a local persistent character array
  char*               GetReferenceBase()     { return ReferenceBase; }
  char*               GetFullReferenceBase();   // get valid FullReferenceBase, building it from ReferenceBase and owner's FullReferenceBase if necessary
  virtual void        FillNetList( TNetList* NetList );
  virtual void        FillDesignatorList( TDesignatorList* DesignatorList );
  virtual void        AssignReferenceDesignators();
  void                CheckReferenceDesignators();   // check for duplicates
  void                WriteReferenceDesignators( ofstream& Dest );  // write   reference designators to file for future use
  int                 ReadReferenceDesignators(  ifstream& Srce );  // read in reference designators from file
  virtual TDesignator SupplyReferenceDesignator( char* aReferencePrefix );
  void                AddReferenceDesignator( TDesignator& D );  // add reference designator to the appropriate designator list
  bool                PortAccessAllowed( TPort* P );  // check that this module is allowed to connect to P
  void                ConnectAll() {   // make local connections and connections in member modules

    // local connections
    ConnectingModule = this;   // give shorthand operators a pointer to the TModule whose Connect() is in progress
    Connect();
    ConnectingModule = 0;      // indicate that no module is connecting--useful for exception handler's message to user

    // connections in members
    for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
      TModule* MP = dynamic_cast<TModule*>( Member );
      if ( MP )  MP->ConnectAll();
    }
  }
  virtual void        Connect() {};    // connections go in here
  virtual void DumpAll( ostream& Dest ) {
    Dest << '\n' << GetTypeName() << ' ' << Name;
    TSubBusList::DumpAll( Dest );
    for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
      Member->DumpAll( Dest );
    }
  }
  void OutputBOM( ofstream& Dest );
//void OutputCad3P( ofstream& Dest, TNetList* NL );  // output netlist in Allegro "third party" format
//void OutputCadPP( ofstream& Dest, TNetList* NL );  // output netlist in PADS PowerPCB format

// wire():
// 1. connect MyPortRange to a new auto-named connection
// 2. for each member module:  connect its port of the specified name (if any) to the same connection
// 3. report an error if MaxMemberCount is exceeded.
// 4. return the number of member ports connected (not including the port of MyPortRange).
// see code for details of default arguments
  int      wire(  TPortRange MyPortRange, char* MemberPortName = 0, int MaxMemberPorts = 1, int LeftRange = -1, int RightRange = -1 );

// wire():
// 1. for each member module:  connect its port of the specified name (if any) to the named connection
// 2. report an error if MaxMemberCount is exceeded.
// 3. return the number of member ports connected.
// see code for details of default arguments
  int      wire(  char* aNamed, char* MemberPortName = 0, int MaxMemberPorts = 1, int LeftRange = -1, int RightRange = -1 );
  int      wire2( char* aNamed, char* MemberPortName = 0, int MaxMemberPorts = 2, int LeftRange = -1, int RightRange = -1 ) {
    return wire(        aNamed,       MemberPortName    ,     MaxMemberPorts    ,     LeftRange     ,     RightRange      );
  }

// wire identically-named (and ranged) ports of Bundle and ModPar together
  int      wire( TBundle& Bundle, TModPar& ModPar );


// wireall():
// shorthand for corresponding wire(), with MaxMemberPorts set to infinite (-1)
  int      wireall( TPortRange MyPortRange, char* MemberPortName = 0,     int LeftRange = -1, int RightRange = -1 )
  { return wire(               MyPortRange,       MemberPortName    , -1,     LeftRange     ,     RightRange      ); }
  int      wireall( char*      aNamed,      char* MemberPortName = 0,     int LeftRange = -1, int RightRange = -1 )
  { return wire(               aNamed,            MemberPortName    , -1,     LeftRange     ,     RightRange      ); }

// merge all wires of Mergee into single wire of Other
  void     merge( TPortRange   Other, TPortRange Mergee );

// merge all wires of Mergee into single wire of aNamed
  void     merge( char*       aNamed, TPortRange Mergee );

// connect an array of ports to named nets
  char* conn( char* Prefix, TPort* Ports, int PortCount );

// auto-named version of conn() has no name specified by user
  TSubBus* conn( const TPortRange& aConn0,
                 const TPortRange& aConn1 = DefaultRange,
                 const TPortRange& aConn2 = DefaultRange,
                 const TPortRange& aConn3 = DefaultRange,
                 const TPortRange& aConn4 = DefaultRange,
                 const TPortRange& aConn5 = DefaultRange,
                 const TPortRange& aConn6 = DefaultRange,
                 const TPortRange& aConn7 = DefaultRange,
                 const TPortRange& aConn8 = DefaultRange,
                 const TPortRange& aConn9 = DefaultRange ) {
      return conn( "", aConn0, aConn1, aConn2, aConn3, aConn4,
                        aConn5, aConn6, aConn7, aConn8, aConn9 );
      }

  TSubBus* conn( char* aName, const TPortRange& aConn0,
                              const TPortRange& aConn1 = DefaultRange,
                              const TPortRange& aConn2 = DefaultRange,
                              const TPortRange& aConn3 = DefaultRange,
                              const TPortRange& aConn4 = DefaultRange,
                              const TPortRange& aConn5 = DefaultRange,
                              const TPortRange& aConn6 = DefaultRange,
                              const TPortRange& aConn7 = DefaultRange,
                              const TPortRange& aConn8 = DefaultRange,
                              const TPortRange& aConn9 = DefaultRange );

  TPortRange bus( char* SubBusName );                                 // return a port range in the specified subbus -- allows operations on the subbus via its name, e.g., associating subbus with a property
  TPortRange bus( char* SubBusName, int LeftRange, int RightRange );  // create a new subbus using an internal port and the specified range
};


class TProperty;     // see sprop.h

class TPropertyList {
private:
  vector<TProperty*>  List;           // "global" list of properties
public:
  void push_back( TProperty* aProperty ) {        List.push_back( aProperty ); }
//unsigned   size()                      { return List.size(); }
  size_t     GetPropertyCount()          { return List.size(); }
  TProperty* GetProperty( unsigned i )   { return List[ i ]; }
  ~TPropertyList();
};



class TPort : public TCBBase, public TRange  {
private:
  bool          mMultiPinNode; // true if multiple pins on single wire
protected:
  TPortType     PortType;      // default = ptUnspecified
  vector<TPin>  Pins;          // Pins are addressed by subtracting MinRange from desired index
                               // 20-Aug-2005 -- Nets and associated code were added/modified to increase performance, e.g., in NetList::Cleanup() and TViewSymbolicNetlist::Dump...()
  vector<int>   Nets;          // index in Root's netlist for each pin (unless multi-pin node), -1 if no net yet assigned

public:
  bool CheckPart();            // true if owner is a TPart
  bool MultiPinNode() { return mMultiPinNode; }

  size_t GetPinCount() { return Pins.size(); }

  void      SetPortType( TPortType aPortType ) {        PortType = aPortType; }
  TPortType GetPortType()                      { return PortType; }

  TPin GetPin( int aIndex ) {
    if ( CheckPart() == false )
      BEGERR << "TPort::GetPin(): owner must be TPart. Port= " << this << ENDERR;
    unsigned i = (unsigned) aIndex;
    if ( ! MultiPinNode() )  i -= MinRange;
    if ( i >= Pins.size() )  return 0;
    return Pins[ i ];
  }

private:
  void InitNets() {                                           // initialize Nets but only if not already initialized
    if ( Nets.size() == 0 ) {                                 // if Nets not yet initialized
      unsigned NetCount = MultiPinNode() ? 1 : Count;         // one net for multi-pin node, else TRange::Count
      for ( unsigned i = 0; i < NetCount; ++ i ) {
        Nets.push_back( -1 );                                 // set initial value to no net
      }
    }
  }

public:                                                       // GetNetIndex() returns index in Root's netlist of specified port pin or -1 if no net yet assigned
  int GetNetIndex( int aIndex ) {                             // this function is intended to be called only after the range of the port is fixed, which typically happens when the port is registered
    InitNets();                                               // make sure Nets was initialized
    unsigned i = MultiPinNode() ? 0 : aIndex - MinRange;      // index within Nets -- all pins of multi-pin node are on same net
    if ( i >= Nets.size() )  return -1;                       // fail silently
    return Nets[ i ];
  }

  void SetNetIndex( int aIndex, unsigned NetIndex ) {         // set the net index for the specified pin
    InitNets();                                               // make sure Nets was initialized
    unsigned i = MultiPinNode() ? 0 : aIndex - MinRange;      // index within Nets -- all pins of multi-pin node are on same net
    if ( i >= Nets.size() )
      BEGERR << "TPort::SetNet(): illegal index of " << aIndex << " for Port " << this << ENDERR;
    Nets[ i ] = (int)( NetIndex );
  }
                                             // functions for setting text for physical package pins
                                             // for each physical pin, there should be exactly one call to one of the functions below
  void SetPin( char* PinName ) {             // for single-wire port
    if ( CheckPart() == false )
      BEGERR << "SetPin() is only allowed for TPart. Port= " << this << ENDERR;
    if ( Pins.size() != 0 )
      BEGERR << "SetPin(): only one pin can be set via SetPin(). Port= " << this << " PinName= " << PinName << ENDERR;
    Pins.push_back( PinName );
#ifdef DEBUG_ON
    cout << "\nSet pin for " << *this << " to " << PinName << '.';
#endif
  }
  void AddPin( char* PinName ) {             // for single-wire port with multiple pins
    if ( CheckPart() == false )
      BEGERR << "AddPin( char*...) is only allowed for TPart. Port= " << this << ENDERR;
    mMultiPinNode = true;
    Pins.push_back( PinName );
#ifdef DEBUG_ON
    cout << "\nAdded pin " << PinName << " to " << *this << '.';
#endif
  }
  void AddPin( int Index, char* PinName ) {    // for multi-wire port
    if ( CheckPart() == false )
      BEGERR << "AddPin( int...) is only allowed for TPart. Port= " << this << ENDERR;
    unsigned i = Index - GetMinRange();        // convert Index into index for Pins[]
    if ( Pins.size() < ( i + 1 ) )  Pins.resize( i + 1 );
    Pins[ i ] = PinName;
#ifdef DEBUG_ON
    cout << "\nAdded pin " << PinName << " to wire " << Index << " of " << *this << '.';
#endif
  }

  // conversion to TPortRange--same as specifying full range of port
  operator TPortRange() { return operator()( LeftRange, RightRange ); }      // full range shorthand, e.g., P

  // overload () to specify port range
  TPortRange operator()( int aLeft ) { return operator()( aLeft, aLeft ); }  // single wire shorthand, e.g., P(4)
  TPortRange operator()( int aLeft, int aRight ) {                           // range shorthand, e.g., P(31,0)
    TPortRange Temp;
    Temp.SetRange( aLeft, aRight );
    if ( ( Temp.GetMinRange() < GetMinRange() ) ||
         ( Temp.GetMaxRange() > GetMaxRange() )     )
         BEGERR << "Bad range requested for port " << this << ". Requested range (" << aLeft << "," << aRight << "). Port's range is (" << GetLeftRange() << "," << GetRightRange() << ")" << ENDERR;
    Temp.SetPort( this );
    return Temp;
  }
  TPort() : Pins( 0 ), Nets( 0 ) {
    PortType   = ptUnspecified;
    mMultiPinNode = false;
  }
  virtual void AddNonWiredPins( TNet* Dest, TNetList* NL );
  virtual void DumpAll( ostream& Dest ) {
    Dest << '\n' << GetTypeName() << ' ' << TPortRange( *this );

    if ( Pins.size() == 0 ) return;
    if ( Pins.size() == 1 ) Dest << "  Pin: ";
    else                    Dest << "  Pins: ";
    if ( MultiPinNode() ) {
      for ( unsigned i = 0; i < GetPinCount(); ++ i ) {
        Dest << ' ' << Pins[ i ];
      }
    }
    else {     // list pins from LeftRange to RightRange
      for ( int i = LeftRange; ; i += ( ( RightRange > LeftRange ) ? 1 : -1 ) ) {
        Dest << ' ' << Pins[ i - MinRange ];
        if ( i == RightRange ) break;
      }
    }
  }
};

// TPort derivatives below are the same as TPort, except PortType is not ptUnspecified
class TInPort : public TPort {
  TInPort() { PortType = ptIn; }
};
class TOutPort : public TPort {
  TOutPort() { PortType = ptOut; }
};
class TInOutPort : public TPort {
  TInOutPort() { PortType = ptInOut; }
};

class TPackage {
public:
  char*       Name;         // e.g., "SOIC16"
  int         PinCount;     // PinCount perhaps only useful for information purposes and double check
  int         GetPinCount() { return PinCount; }
  char*       GetName()     { return Name;     }
  TPackage() { Name = 0; PinCount = -1; }
};


class TPart : public TModPar {
  TDesignator  ReferenceDesignator;    // reference designator
  char*        ReferencePrefix;        // default prefix for reference designator, e.g., U
  TPackage     Package;                // package type

public:
  TPart() { ReferencePrefix = 0; }
  void SetReferencePrefix( char* aReferencePrefix ) {
    ReferencePrefix = aReferencePrefix;
  }
  void SetPackage( char* aName, int aPinCount = -1 ) {
    Package.Name = aName;                 // package details TBD
    if ( aPinCount != -1 )  Package.PinCount = aPinCount;
  }
  TPackage    GetPackage() {
    if ( Package.GetName() == 0 )
      BEGERR << "Part " << *this << " has no package. Each part class must assign a package via TPart::SetPackage()" << ENDERR;
    return Package;
  }
  TDesignator GetReferenceDesignator() { return ReferenceDesignator; }
  virtual void        AssignReferenceDesignators();
  void                AssignReferenceDesignator( TDesignator D );
  virtual TDesignator SupplyReferenceDesignator( char* aReferencePrefix );
  virtual void        FillDesignatorList( TDesignatorList* DesignatorList );


  virtual void DumpAll( ostream& Dest ) {
    Dest << '\n' << GetTypeName() << ' ' << Name;
    for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
      Member->DumpAll( Dest );
    }
  }

};


// TBundle is the base class for bundles of TPort's and TBundles
// It is used for creating complex busses.

class TBundle : public TModPar {

public:
  TBundle() {}
  // TBundle must only contain TPort and TBundle descendants
  virtual void        AssignReferenceDesignators() {};  // TBundle has no member TPart's that would need reference designators
  virtual TDesignator SupplyReferenceDesignator( char* aReferencePrefix ) {
    BEGERR << this << "::SupplyReferenceDesignator() should never be called" << ENDERR;
  }
  virtual void DumpAll( ostream& Dest ) {
    Dest << '\n' << GetTypeName() << ' ' << Name;
    for ( TCBBase* Member = FirstMember; Member != 0; Member = Member->Next ) {
      Member->DumpAll( Dest );
    }
  }
};


class TPinSpec {  // pin specifier
  TPort* Port;
  int    Index;   // index within port
public:
  bool   OwnerIsPart()   // return true if Port's Owner is a TPart
    { return ( dynamic_cast<TPart*>( Port->GetOwner() ) != 0 ); }
  TPort* GetPort()  { return Port; }
  size_t GetPinCount() {
    if ( MultiPinNode() )         return Port->GetPinCount();
    if ( Port->GetCount() == 0 )  return 0;
                                  return 1;
  }
  bool   MultiPinNode() { return Port->MultiPinNode(); }
  int    GetIndex() { return Index; }
  TPin   GetPin( unsigned aIndex ) {
    if ( ! MultiPinNode() )  return Port->GetPin( Index  );  // ignore user-supplied index
                             return Port->GetPin( aIndex );
  }

  int    GetNetIndex() {                        // get index of associated net as recorded in Port
    return Port->GetNetIndex( Index );
  }

  void   SetNetIndex( unsigned NetIndex ) {     // set the net index in Port
    Port->SetNetIndex( Index, NetIndex );
  }


  TPinSpec() { Port = 0; Index = -1; }
  TPinSpec( TPort* aPort, int aIndex ) { Port = aPort; Index = aIndex;}
  bool operator==( TPinSpec& aPinSpec ) {
    return ( ( aPinSpec.Port  == Port  ) &&
             ( aPinSpec.Index == Index )    );
  }

  ostream& OutputPins( ostream& Dest, bool Brief = false ) {  // output pin(s) to Dest
  if ( Port->CheckPart() == false )
    BEGERR << "TPinSpec::OutputPins(): Port's owner must be a TPart. Port= " << Port << ENDERR;

    if ( Port->MultiPinNode() ) {
      if ( Brief ) {                                 // brief mode--output one pin only
        Dest << Port->GetPin( 0 );
        if ( Port->GetPinCount() > 0 ) Dest << "...";  // indicate additional pins
      }
      else {
        for ( unsigned i = 0; i < Port->GetPinCount(); ++ i ) {
          if ( i != 0 )  Dest << ' ';
          Dest << Port->GetPin( i );
        }
      }
    }
    else {
      Dest << Port->GetPin( Index );
    }
    return Dest;
  }
};


bool operator<( TPinSpec A, TPinSpec B );      // use with stdlib's sort()


class TNetName : public TDesignator {
  virtual void Dummy() {}  // to make polymorphic
  int Extension;           // used to generate unique net names when base TDesignator is not unique within a netlist
public:
  TNetName() { Extension = 0; }
  int  GetExtension()                 { return Extension; }
  void SetExtension( int aExtension ) { Extension = aExtension; }
};


class TNet : public TNetName {   // a net is a single wire with one or more pins attached to it
  unsigned PartPinSpecCount;     // count of PinSpecs whose port is owned by a TPart
public:
  vector<TPinSpec> PinSpecs;
  TNet() { Clear(); }
  void Clear() { PinSpecs.clear(); PartPinSpecCount = 0; }

  size_t  GetPinSpecCount() { return PinSpecs.size(); }
  unsigned  GetPartPinSpecCount() { return PartPinSpecCount; }

  TPinSpec* GetPinSpec( unsigned p ) {
    if ( p >= PinSpecs.size() )
      BEGERR << "GetPinSpec( " << p << " ):  vector index limit of " << PinSpecs.size() << " exceeded for net " << this << ENDERR;
    return &PinSpecs[ p ];
  }

  void DumpAsPortRanges( ostream& Dest, char* Separator = "\n  " );    // output net as list of PortRange's (rather than list of PinSpec's )

  void AddPin( TPinSpec aPinSpec ) {
    if ( PinInNet( aPinSpec ) ) return;    // suppress duplicates--they are a normal byproduct of the netlist building algorithm
    PinSpecs.push_back( aPinSpec );
    if ( dynamic_cast<TPart*> ( aPinSpec.GetPort()->GetOwner() ) )  PartPinSpecCount ++;
#ifdef DEBUG_ON
  cout << "\nAdded pin " << aPinSpec << " to net " << TNetName( *this ) << '.';
#endif
  }

  void       AddPin( TPort* aPort, int aIndex ) { AddPin( TPinSpec( aPort, aIndex ) ); }
  void       AddNet( TNet* aNet );            // add pins from aNet (except duplicates)
  TPinSpec*  CommonPins( TNet* aNet );        // return PinSpec of pin common to both this net and aNet
  bool       PinInNet( TPinSpec aPinSpec ) {  // true if aPinSpec is in net
    for ( unsigned i = 0; i < PinSpecs.size(); ++ i ) {
      if ( PinSpecs[ i ] == aPinSpec )  return true;
    }
    return false;
  }

  void SetNetIndex( unsigned Index ) {
    for ( unsigned i = 0; i < PinSpecs.size(); ++ i ) {
      PinSpecs[ i ].SetNetIndex( Index );
    }
  }

                                              // return true if any PinSpec's net index is changed
  bool SetMinNetIndex( unsigned Index ) {     // set each PinSpec's net index to the minimum net index within net, or Index, whichever is smaller
    bool AnyIndexChanged = false;             // return value
    size_t PS_Count = PinSpecs.size();

    unsigned MinI = PinSpecs[ 0 ].GetNetIndex();
    if ( Index < MinI )  MinI = Index;

    for ( size_t i = 0; i < PS_Count; ++ i ) {    // find minimum index
      int CurrentI = PinSpecs[ i ].GetNetIndex();
      AnyIndexChanged = AnyIndexChanged || ( MinI != (unsigned)( CurrentI ) );
      if ( (unsigned)( CurrentI ) < MinI )   MinI  = (unsigned)( CurrentI )  ;
    }

    if ( AnyIndexChanged ) {   // if any index changed, then set all PinSpecs to use the smallest index
//    cout << "\nIndex " << Index << " net index changed to " << MinI << ".";
      for ( unsigned i = 0; i < PS_Count; ++ i ) {
        PinSpecs[ i ].SetNetIndex( MinI );
      }
    }

    return AnyIndexChanged;
  }

  int GetNetIndex() {    // get net index as recorded in first PinSpec's Port
    if ( PinSpecs.size() == 0 )  return -1;
    return  PinSpecs[ 0 ].GetNetIndex();
  }

  void Sort() {          // sort PinSpec's using operator<()
    sort( PinSpecs.begin(), PinSpecs.end() );  // standard library sort
  }
};


class TNetList {
private:
  bool mCleared;
  bool mCleaned;         // true --> netlist has been cleaned up --> net indexes in PinSpec can be used to identify net containing PinSpec
  vector<TNet> Nets;
  
public:

  TNet* NewNet() { Nets.push_back( TNet() );  return &Nets.back(); }

  size_t GetNetCount() { return Nets.size(); }
  TNet*    GetNet( unsigned n ) {
    if ( n >= Nets.size() )
      BEGERR << "TNetList::GetNet( " << n << " ):  vector index limit of " << Nets.size() << " exceeded" << ENDERR;
    return &Nets[ n ];
  }
  void Clear()   { mCleared = true; Nets.clear(); }
  bool Cleared() { return mCleared; }
  bool Cleaned() { return mCleaned; }
  void AddSubBus( TSubBus& SubBus );                                      // Add SubBus to netlist without merging wires into existing nets
  void AddSubBusList( TSubBusList& SubBusList, char* ExcludeName = "" );  // call AddSubBus() for each SubBus in SubBusList unless it is named ExcludeName.
  TNet* PinInNetList( TPinSpec aPinSpec ) {              // return pointer to first net that contains aPinSpec, or 0 if none
    if ( Cleaned() ) {
      int  NetIndex = aPinSpec.GetNetIndex();
      if ( NetIndex == -1 )  return 0;         // if user did not connect the PinSpec
      return GetNet( ( unsigned ) NetIndex );  // get net using index -- GetNet() handles illegal indexes
    }
    for ( unsigned n = 0; n < Nets.size(); ++ n ) {
      for ( unsigned p = 0; p < Nets[ n ].PinSpecs.size(); ++ p ) {
        if ( Nets[ n ].PinSpecs[ p ] == aPinSpec )  return &Nets[ n ];
      }
    }
    return 0;
  }

  void SetNetIndexes() {
    for ( unsigned n = 0; n < Nets.size(); ++ n ) {
      Nets[ n ].SetNetIndex( n );
    }
  }

  void Cleanup( bool Sort = true );
  
  void SortPinSpecs() {       // sort the PinSpec's in each net -- the order of nets in Nets is not changed
    for ( unsigned n = 0; n < Nets.size(); ++ n ) {
      Nets[ n ].Sort();                               // sort PinSpec's by registration numbers of Ports and, if same port, then by index within port
    }
  }

  void SortNets( bool ReIndex = true ) {              // sort nets alphabetically by name -- use false ReIndex to preserve existing net indexes stored in TPort's -- e.g., see TNetList::DumpSorted()
    sort( Nets.begin(), Nets.end() );                 // standard library sort() -- TNet's are derived from TNetName -- sort is based on operator<( TNetName, TNetName )
    if ( ReIndex ) SetNetIndexes();                   // new net indexes are typically needed after sort
  }
  
  TNetList() { Clear(); mCleaned = false; }
  void Dump(               ostream& Dest );
  int  DumpSingleNodeNets( ostream& Dest );
  int  DumpNoConnectNet(   ostream& Dest );
  void DumpSorted(         ostream& Dest, bool DumpAll = true );
};


class TLogStatus {     // class used only to output progress info to user
private:
  time_t    CreationTime;     // time class instance was created
  time_t    StartTime;        // time of previous call to Log
  ostream*  Dest;
  bool      FirstLog;
  bool      mEnable;
public:
  TLogStatus( ostream& aDest ) {
    Dest = &aDest;
    CreationTime = time( 0 );
    StartTime = CreationTime;
    FirstLog = true;
    mEnable = true;
  }
  void Log( char* StartString, bool Ellipsis = true ) {
    if ( mEnable ) {
      if ( ! FirstLog ) {             // if not first call, then output conclusion for previous call
        *Dest << ". Elapsed time: " << ( time( 0 ) - StartTime ) << " s.";
      }
      *Dest << '\n' << StartString;
      if ( Ellipsis ) {
        *Dest << "...";
      }
      *Dest << flush;
    }
    StartTime = time( 0 );
    FirstLog = false;
  }
  void End() {
    if ( ! FirstLog )  Log( "", false );  // conclude any previous output
    if ( mEnable ) {
      *Dest << "Total elapsed time: " << ( time( 0 ) - CreationTime ) << " s.\n" << flush;
    }
  }
  void Enable( bool aEnable = true ) {
    mEnable = aEnable;
  }
};

void operator<<( TLogStatus& L, char* aString );

char* DateTimeString();


class TPath {
public:
  vector<TModPar*>  ModPars;   // list of modules in path, starting with oldest ancestor (excluding root)

  void Set( TModPar* MP ) {    // set path to be path of specified ModPar
    ModPars.clear();
    if ( ! MP )  return;
    // ancestors are listed oldest first
    int        AncestorCount = 0;
    TModPar*   Ancestor      = MP->GetOwner();
    for ( ; Ancestor != 0; AncestorCount ++ ) { Ancestor = Ancestor->GetOwner(); }

//    if ( AncestorCount != 0 ) AncestorCount--;   // in most cases, do not list root
    for( int i = AncestorCount; i != 0; -- i ) {
      Ancestor = MP->GetOwner();
      for ( int j = 0; j < (i-1); ++ j ) Ancestor = Ancestor->GetOwner();   // seek ancestor
      ModPars.push_back( Ancestor );
    }
    ModPars.push_back( MP );
  }

  void Clear() {
    ModPars.clear();
  }

  TPath(){};
  TPath( TModPar* MP ) { Set( MP ); }
};



// The two classes below were added in order to eliminate the former static global module that contained the subbus list for global connections (e.g., /NC)
// The new code allows a single program to contain more than one root module, e.g., if a single program is used to output netlists for different versions of a single design (e.g., backplanes that are identical except for the number of slots).

class TGlobalSubBusModule : public  TModule {  // class used only by TRootModule, below, to hold global connections
  virtual void Register() {}                   // required to make non-abstract
};

                                            // >>> a design's root module must be derived from TRootModule
class TRootModule : public TModule {        // module at the root of a hierarchy
private:
  int                  NextSerialNumber;    // serial number for next TCBBase registered
public:
  TGlobalSubBusModule  GlobalSubBusModule;  // holds the global subbus list for this hierarchy
  TNetList             NetList;             // holds the netlist for this hierarchy
  bool                 VerboseNetList;      // true --> NetList (above) will contain all PinSpecs, else only PinSpecs for Part Ports:
                                            // For large designs, setting VerboseNetList to false can significantly decrease execution time.
                                            // Most postprocessing requires NetList to contain only PinSpecs for Part Ports.
                                            // Some postprocessing, such as user property interpreters, can benefit from the additional Module Port information (including Ports of Bundles) available in NetList when VerboseNetList is true.
  TPropertyList        Properties;          // list of all properties created within this hierarchy

  TRootModule() {
    VerboseNetList          = true;
    GlobalSubBusModule.Name.SetDesignator( "GlobalSubBusModule" );
    Root                    = this;
    GlobalSubBusModule.Root = this;
    NextSerialNumber        = 0;
  }

  void ConnectAll() {
    TModule::ConnectAll();     // --> recursive call of all Connect()'s
    FillNetList();             // --> recursive call of all FillNetLists()'s
  }

  virtual void  FillNetList( TNetList* aNetList = 0 ) {   // add subbusses to netlist
    TNetList* NL = aNetList ? aNetList : &NetList;        // typically, aNetList is 0 -- default value of aNetList was added mainly to avoid warning (hiding virtual function)
    TModule::FillNetList( NL );
    NL->Cleanup();
  }

  void   DumpNetList(        ostream& Dest               ) {        NetList.Dump(               Dest          ); }
  void   DumpSortedNetList(  ostream& Dest, bool DumpAll ) {        NetList.DumpSorted(         Dest, DumpAll ); }
  int    DumpSingleNodeNets( ostream& Dest               ) { return NetList.DumpSingleNodeNets( Dest          ); }
  int    DumpNoConnectNet(   ostream& Dest               ) { return NetList.DumpNoConnectNet(   Dest          ); }
  size_t GetNetCount()                                     { return NetList.GetNetCount(); }
  int    GetSerialNumber()                                 { return NextSerialNumber ++; }
  int    GetNextSerialNumber()                             { return NextSerialNumber;    }   // e.g., can be used after registration pass to determine number of TCBBase's registered

  virtual void  AddNonWiredPins( TNet* DestNet, TNetList* aNetList = 0 ) {  // add non-wired pins to DestNet, e.g., for reporting non-wired pins
    TNetList* NL = aNetList ? aNetList : &NetList;        // typically, aNetList is 0 -- default value of aNetList was added mainly to avoid warning (hiding virtual function)
    TModule::AddNonWiredPins( DestNet, NL );
  }

  void  DumpGlobals( ostream& Dest ) {
    Dest << "\nGlobals:";
    GlobalSubBusModule.DumpAll( Dest );
    Dest << '\n';
  }

};

#endif

