/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_char_bucket.h:  write-only character string container

#ifndef _cb_char_bucketH_
#define _cb_char_bucketH_

// Use AddString() to add a character string to the bucket.
// AddString() returns a char* that is valid until the CharBucket is destroyed.

#include <vector>
#include <string>


class TCharBucket {
private:
  size_t   MaxLength;      // maximum length of one string (including terminating null)--allocate memory MaxLength chars at a time
  vector<char*> Arrays;    // pointers to those arrays (needed only for memory deallocation)
  char* CurrentArray;      // pointer to current destination array
  size_t CurrentIndex;   // index within CurrentArray[]
public:
  TCharBucket( size_t aMaxLength = 4096 ) {
    MaxLength = aMaxLength;
    CurrentArray = 0;
    CurrentIndex = MaxLength;   // force allocation of memory on first call of AddString()
  }
  ~TCharBucket() {
    for ( size_t i = 0; i < Arrays.size(); ++ i ) {
      delete[]Arrays[ i ];
    }
  }
  char* AddString( const char* aString ) {
    size_t Length = strlen( aString ) + 1;          // length including terminating null
    if ( Length > MaxLength )
      throw( "TCharBucket: string length exceeded." );   // !! throwing a char*

    if ( CurrentIndex + Length > MaxLength ) {
      CurrentArray = new( char[ MaxLength ] );   // allocate new array
      Arrays.push_back( CurrentArray );          // keep track of where it is -- needed only for deallocation
      CurrentIndex = 0;
    }
    char* Dest = CurrentArray + CurrentIndex;
    strcpy( Dest, aString );
    CurrentIndex += Length;
    return Dest;
  }
};


#endif

