/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_netlist_pp.cpp:  netlist generator for PADS layout tool (PADS is a registered trademark of Mentor Graphics Corporation.)

#include "cb_base.h"
#include "cb_shorthand.h"


// output standard netlist format (PADS PowerPCB V4.0 format)
void OutputCadPP( TRootModule* Root, ofstream& Dest, bool NullTimeStamp ) {
  TNetList* NL = &Root->NetList;
  Dest <<   "!CREATED-BY-CBOLD! DESIGN DATABASE ASCII FILE 1.0";
  if ( NullTimeStamp )  Dest << "\n!" << "Sat Jan 01 00:00:00 2000" << "!";
  else                  Dest << "\n!" <<  DateTimeString()          << "!";


// *** output part list ***//
  Dest << "\n*PART*";

// format is:
// ref_desig   part_type@layout_symbol
// U10            AD1580@SOT23

  const int Indent = 12;
  ostringstream RefDes;     // accepts reference designator, used mainly to determine number of characters in reference designator
  for ( TCBBase* CBB = Root->IterFirst(); CBB != 0; CBB = Root->IterNext() ) {
    TPart* Part = dynamic_cast<TPart*>( CBB );
    if ( Part ) {
      Dest << "\n" << Part->GetReferenceDesignator();
      RefDes.str("");                              // clear the string
      RefDes << Part->GetReferenceDesignator();    // this code is necessary only for justifying text following reference designator
      int Spaces = Indent - (int)(RefDes.str().length());
      if ( Spaces < 2 )  Spaces = 2;
      for ( int s = 0; s < Spaces; ++ s )  Dest << ' ';
      Dest << Part->GetTypeName();
      Dest << '@';
      Dest << Part->GetPackage();
    }
  }




// *** output net list ***//
  Dest << "\n*NET*";
  ostringstream Net;         // accepts net name, used mainly to replace GlobalChar
  string NetString;

  for ( unsigned n = 0; n < NL->GetNetCount(); ++ n ) {   // for each net
    TNet* N         = NL->GetNet( n );
    if ( N->DesignatorEquals( NoConnect ) )  continue;    // discard no-connect net

    Net.str( "" );                    // clear string
    Net << *N;                        // get net name as string
    NetString = Net.str();
// *** use the code below to replace GlobalChar with another char, e.g., an underscore ***
// *** user must guarantee that this does not create two nets with identical names
    if ( NetString[ 0 ]  ==  GlobalChar )  NetString[ 0 ] = '_';

    Dest << "\n*SIGNAL* " << NetString << "\n";

    const int MaxPinsOnLine = 8;  // max pins to output on a single line
    int PinsOnLine = 0;
    char* Separator = "";

    for ( unsigned p = 0; p < N->GetPinSpecCount(); ++ p ) {
      TPinSpec PS = *N->GetPinSpec( p );

      TPart* OP = dynamic_cast<TPart*>( PS.GetPort()->GetOwner() );
      if ( OP == 0 )  continue;  // as of 14-Aug-2005, TNet's contain all ports, not just ports of parts
//    if ( OP == 0 )
//      BEGERR << "OutputCadPP(): port " << PS.GetPort() << " is owned by non-Part" << ENDERR;
      TDesignator Reference = OP->GetReferenceDesignator();

      size_t PinCount = PS.GetPinCount();
      if ( PinCount == 0 )
        BEGERR << "TModule::OutputCadPP():  PinSpec " << PS << " has no physical pin" << ENDERR;

      for ( unsigned i = 0; i < PinCount; ++ i ) {
        Separator = " ";
        if ( PinsOnLine >= MaxPinsOnLine ) {
          PinsOnLine = 0;
          Dest << "\n";   // add newline
        }
        if ( PinsOnLine == 0 )  Separator = "";                    // use space as separator except at beginning of line
        Dest << Separator << Reference << '.' << PS.GetPin( i );   // i is ignored except for multipin node
        PinsOnLine++;
      }
    }
  }
/* remainder of file has form:
*MISC*
ATTRIBUTE VALUES
{
PART R11
{
"Value" 22K
}
}
*END*
*/


  Dest << "\n\n*MISC*";
  Dest << "\nATTRIBUTE VALUES";
  Dest << "\n{";

// for each part
  for ( TCBBase* CBB = Root->IterFirst(); CBB != 0; CBB = Root->IterNext() ) {
    TPart* Part = dynamic_cast<TPart*>( CBB );
    if ( Part ) {
      Dest << "\nPART " << Part->GetReferenceDesignator();
      Dest << "\n{";
      Dest << "\n\"Value\" " << Part->GetTypeName();
      Dest << "\n}";
    }
  }

  Dest << "\n}";            // end of ATTRIBUTE VALUES
  Dest << "\n*END*";        // end of file
}

