/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_prop.h:  classes related to properties -- see notes at the end of this file

#ifndef _cb_propH_
#define _cb_propH_

#include "cb_base.h"

#define property( atype, aref ) atype& aref = *(new atype)

ostream& operator<<( ostream& Dest, TProperty& P );

class TProperty {  // base class for all properties
private:
  TModPar*                   Owner;         // owner (0 if property was created outside of Connect())
  vector<TModPar*>           ModPars;
  vector<TPortRange>         PortRanges;
  vector<TBundle*>           Bundles;
  vector<TSubBusSpec>        SubBusses;
protected:
  void associate( TModPar&     aModPar          ) {    ModPars.push_back( &aModPar          ); }
  void associate( TPortRange&  aPortRange       ) { PortRanges.push_back( aPortRange        ); }
  void associate( TBundle&     aBundle          ) {    Bundles.push_back( &aBundle          ); }
  void associate( char* aName, TModule* aModule ) {  SubBusses.push_back( TSubBusSpec( aName, aModule ) ); }

  void illegal(   TModPar&    aModPar    ) {
    BEGERR << "Illegal attempt to associate property " << this << " with " << aModPar << ENDERR;
  }
  void illegal(   TPortRange& aPortRange ) {
    BEGERR << "Illegal attempt to associate property " << this << " with " << aPortRange << ENDERR;
  }
  void illegal(   TBundle& aBundle ) {
    BEGERR << "Illegal attempt to associate property " << this << " with " << aBundle << ENDERR;
  }
  void illegal(   char* aName, TModule* aModule ) {
    BEGERR << "Illegal attempt to associate property " << this << " with SubBus " << aName << " in module " << aModule << ENDERR;
  }

public:
  TProperty() {                    Owner = TModPar::RegisteringModPar;
                if ( Owner == 0 )  Owner = TModule::ConnectingModule;
                if ( Owner == 0 )  Owner = TModule::PostConnectingModPar;
                if ( Owner == 0 ) {
                  BEGERR << "A property may only be instantiated within Register(), Connect() or PostConnect()" << ENDERR;
                }
                Owner->Root->Properties.push_back( this );
              }

  TModPar*    GetOwner() { return Owner; }
  size_t      GetModParCount()               { return ModPars.size();    }
  TModPar*    GetModPar(    unsigned i )     { return ModPars[ i ];      }
  size_t      GetPortRangeCount()            { return PortRanges.size(); }
  TPortRange  GetPortRange( unsigned i )     { return PortRanges[ i ];   }

// derived classes must have an Associate() for each type of object with which the class can be associated -- these default Associate()'s handle those types with which a class may not be legally associated                                                                                  
  virtual void Associate(     TModPar& aModPar           ) { illegal( aModPar        ); }
  virtual void Associate(  TPortRange& aPortRange        ) { illegal( aPortRange     ); }
  virtual void Associate( TBundle& aBundle       ) { illegal( aBundle    ); }
  virtual void Associate( char* aName, TModule* aModule  ) { illegal( aName, aModule ); }

// prevent user from doing something with unintended results
// making copying illegal also spares derived classes the burden of coding the following functions
  TProperty(            const TProperty& P ) { CBOLD_CopyError( "TProperty", "construction" );               }   // copy constructor
  TProperty& operator=( const TProperty& P ) { CBOLD_CopyError( "TProperty", "assignment"   ); return *this; }   // copy assignment
};

TProperty& operator<=( TProperty& P, TModPar& ModPar );
TProperty& operator<=( TProperty& P, TPortRange PortRange );
TProperty& operator<=( TProperty& P, TBundle& Bundle );

// associate the property with the named SubBus in the currently connecting module
// this operator may only be used within the Connect() of the module that contains the named Subbus
// the SubBus must already exist
TProperty& operator<=( TProperty& P, char* SubBusName );

TModPar&   operator<=( TModPar&   ModPar,    TProperty& P );
TPortRange operator<=( TPortRange PortRange, TProperty& P );
TBundle&   operator<=( TBundle&   Bundle,    TProperty& P );

// associate the property with the named SubBus in the currently connecting module
// this operator may only be used within the Connect() of the module that contains the named Subbus
// the SubBus must already exist
char*      operator<=( char* SubBusName, TProperty& P );


/*

Properties are carriers of information that can be associated with certain CBOLD objects.
A single property instance can be associated with multiple CBOLD objects.
Properties must be persistent -- i.e., they should be created with new and never deleted.
When a property is created, it is automatically added to the global (root) PropertyList.

Properties do very little besides carry information. Their use is optional.
Properties are typically used by post-processing code, e.g., to create layout script files.


Properties can be associated with objects of type:
  TModPar (i.e., TModule or TPart)
  TPortRange
  TBundle
  TSubBus (?)

  For TSubBus, the user specifies the SubBus by name, and operator<=() creates a
  TSubBusSpec that holds pointers to the name and the current connecting module.
  Alternatively, the user may call Associate() and specify the SubBus name and
  module directly.



It is not legal to associate a property with a SubBus that has not yet been created.
The only thing that would be gained by making it legal is the (dubious)
convenience of being able to associate a property with a SubBus before the SubBus
has been named.  More likely than not, this would indicate that the designer
did something he did not intend to do.


Example use of TProperty:

  property( CR_MatchedClock, P );          // create new property and pointer to it -- assumes CR_MatchedClock was derived from TProperty

  CR_MatchedClock Match1 = new CR_MatchedClock();    // equivalent code

  TProperty P( "MatchedClock", "Group1" );
  P <= ClockDriver1.OUT( 11, 0 );            // associate port range with P
  P <= ClockDriver2;                         // associate part with P
  P <= "MyNet"      // a subbus


*/

#endif

