/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_shorthand.cpp:  shorthand operators for making connections

#include "cb_shorthand.h"


TPortRange operator<<( TPortRange PortRangeA, TPortRange PortRangeB ) {
  TModule::ConnectingModule->conn( PortRangeA, PortRangeB );
  return PortRangeA;
}

char* operator<<( char* CP, TPortRange PortRange  ) {
  TModule::ConnectingModule->conn( CP, PortRange );
  return CP;
}

char* operator<<( TPortRange PortRange, char* CP ) {
  return CP << PortRange;
}


// **** Operators for connecting ports of TBundle's **** //

// operator for connecting TBundle's ports to named nets
// the net names are derived from the user-specified prefix string as well as port names and sub-bundle names
// the actual character strings are maintained in a permanent string container

// operator<<( const char* Prefix, TBundle& Bundle ) does not return a char* in order to prevent the user from doing something like this:
// "PREFIX_" << MyBundle << MyPort;
// If the operator had returned a char*, the above statement would be legal and equivalent to:
// "PREFIX_" << MyBundle;    // use "PREFIX_" as connection prefix -- ok
// "PREFIX_" << MyPort;      // use "PREFIX_" as connection name -- most likely not what user thought he was doing

// it is legal for Prefix to begin with GlobalChar

TBundle& operator<<( const char* Prefix, TBundle& Bundle ) {
  string NetPrefix = Prefix;       // initial prefix--this will grow if there are embedded TBundle's

  if ( Bundle.Name.GetNumber() != -1 ) {                              // if bundle is part of an array, append index to net prefix
    NetPrefix += SCSTRING( Bundle.Name.GetNumber() );                 // convert number to string
    NetPrefix +=  "_";                                                // separate array index from remainder of net name
  }

  for ( TCBBase* Member = Bundle.GetFirstMember(); Member != 0; Member = Member->Next ) {
    TPort* Port                 = dynamic_cast<TPort*>(       Member );
    TBundle* SubBundle  = dynamic_cast<TBundle*>( Member );
    if ( Port ) {
      string NetSuffix;  // empty string
      if ( Port->Name.GetNumber() != -1 ) {                     // if port is part of an array, append index to net name
        NetSuffix =  SCSTRING( Port->Name.GetNumber() );        // convert number to string
        if ( Port->GetCount() > 1 )  NetSuffix += "_";          // separate array index from wire index
      }
      string NetName    = NetPrefix + Port->Name.GetString() + NetSuffix;
      char*  NetNamePtr = PermString( NetName.c_str() );  // pointer to permanent null-terminated char array

      #ifdef UppercaseNetNames                            // use preprocessor to avoid "Condition is always true" compiler warning
        for ( int i = 0; NetNamePtr[ i ] != 0; ++ i ) {
          NetNamePtr[ i ] = toupper( NetNamePtr[ i ] );   // convert name to uppercase
        }
      #endif

      TModule::ConnectingModule->conn( NetNamePtr, *Port );
    }
    else if ( SubBundle ) {  // descend recursively into sub-bundles
      string SubPrefix = NetPrefix + SubBundle->Name.GetString() + "_";    // form new net name prefix
      SubPrefix.c_str() << *SubBundle;                     // recursive call to operator<<( const char* Prefix, TBundle& Bundle )
    }
    else
      BEGERR << "operator<<( char*, TBundle& ): Unexpected type encountered as member of TBundle descendant " << Bundle << ENDERR;
  }

  return Bundle;
}




// operator for connecting TBundle's of identical type

TBundle& operator<<( TBundle& BundleA, TBundle& BundleB ) {
  if ( typeid( BundleA ) != typeid( BundleB ) )
    BEGERR << "operator<<( TBundle&, TBundle& ): TBundle descendants " << BundleA << " and " << BundleB << " must be of the same type" << ENDERR;

  TCBBase* MemberA = BundleA.GetFirstMember();
  TCBBase* MemberB = BundleB.GetFirstMember();
  if ( ! MemberA  ||  ! MemberB )
    BEGERR << "operator<<( TBundle&, TBundle& ): Bundles must be registered and non-empty"   << ENDERR;
  while ( MemberA || MemberB ) {
    if ( typeid( MemberA ) != typeid( MemberB ) )
      BEGERR << "operator<<( TBundle&, TBundle& ): Members of TBundle descendants " << BundleA << " and " << BundleB << " are inconsistent" << ENDERR; // should never happen
    TPort* PortA                = dynamic_cast<TPort*>(       MemberA );
    TPort* PortB                = dynamic_cast<TPort*>(       MemberB );
    TBundle* SubBundleA = dynamic_cast<TBundle*>( MemberA );
    TBundle* SubBundleB = dynamic_cast<TBundle*>( MemberB );
    if ( PortA || PortB ) {
      if ( ! PortA  ||  ! PortB )
        BEGERR << "operator<<( TBundle&, TBundle& ): Bundle member ports must be registered" << ENDERR;
      TModule::ConnectingModule->conn( *PortA, *PortB );
    }
    else if ( SubBundleA || SubBundleB ) {  // descend recursively into sub-bundles
      if ( ! SubBundleA  ||  ! SubBundleB )
        BEGERR << "operator<<( TBundle&, TBundle& ): Subbundles must be registered" << ENDERR;
      *SubBundleA << *SubBundleB;
    }
    else
      BEGERR << "operator<<( TBundle&, TBundle& ): Unexpected types encountered as members of TBundle descendants " << BundleA << " and " << BundleB << ENDERR;

    MemberA = MemberA->Next;
    MemberB = MemberB->Next;
  }

  return BundleA;
}




// **** Operators that accept a TModPar **** //
// When the TModPar is on the right of the ^, it must have one or two ports; its first port is connected.
// When the TModPar is on the left of the ^, it must have two ports; its second port is connected.
// operator^() is used rather than operator<<() because:
//  1. the separate syntax emphasizes that a part rather than a port is the operand
//  2. operator^() has lower precedence than operator<<(), e.g.,
//     "FOO" << Fred.Out   ^  Resistor1 ^   Bill.In << "BAR"  ;
//          is equivalenet to:
//   ( "FOO" << Fred.Out ) ^  Resistor1 ^ ( Bill.In << "BAR" );


TModPar& operator^( TPortRange PortRange, TModPar& ModPar ) {
  int PortCount = ModPar.GetPortCount();
  if ( PortCount != 1  &&  PortCount != 2 )
    BEGERR << "operator^( " << PortRange << ", " << ModPar << " ): illegal port count of " << PortCount << ENDERR;

  TModule::ConnectingModule->conn( PortRange, *ModPar.GetPort( 0 ) );  // connect PortRange to ModPar's first port
  return ModPar;
}

TModPar& operator^( char* CP, TModPar& ModPar ) {
  int PortCount = ModPar.GetPortCount();
  if ( PortCount != 1  &&  PortCount != 2 )
    BEGERR << "operator^( \"" << CP << "\", " << ModPar << " ): illegal port count of " << PortCount << ENDERR;

  TModule::ConnectingModule->conn( CP, *ModPar.GetPort( 0 ) );  // connect PortRange to ModPar's first port
  return ModPar;
}


TModPar& operator^( TModPar& ModPar_L, TModPar& ModPar_R ) {
  int PortCount = ModPar_L.GetPortCount();
  if ( PortCount != 2 )
    BEGERR << "operator^( " << ModPar_L << ", " << ModPar_R << " ): illegal port count of " << PortCount << " for " << ModPar_L << ENDERR;

  PortCount = ModPar_R.GetPortCount();
  if ( PortCount != 1  &&  PortCount != 2 )
    BEGERR << "operator^( " << ModPar_L << ", " << ModPar_R << " ): illegal port count of " << PortCount << " for " << ModPar_R << ENDERR;

  TModule::ConnectingModule->conn( *ModPar_L.GetPort( 1 ), *ModPar_R.GetPort( 0 ) );  // connect ports of two parts
  return ModPar_R;
}


char* operator^( TModPar& ModPar, char* CP ) {
  int PortCount = ModPar.GetPortCount();
  if ( PortCount != 2 )
    BEGERR << "operator^( " << ModPar << ", \"" << CP << "\" ): illegal port count of " << PortCount << ENDERR;

  TModule::ConnectingModule->conn( CP, *ModPar.GetPort( 1 ) );  // connect PortRange to ModPar's second port
  return CP;
}


TPortRange operator^( TModPar& ModPar, TPortRange PortRange ) {
  int PortCount = ModPar.GetPortCount();
  if ( PortCount != 2 )
    BEGERR << "operator^( " << ModPar << ", " << PortRange << " ): illegal port count of " << PortCount << ENDERR;

  TModule::ConnectingModule->conn( PortRange, *ModPar.GetPort( 1 ) );  // connect PortRange to ModPar's second port
  return PortRange;
}

