/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_views.cpp:  symbolic netlist generator

#include "cb_views.h"

// dump Module to Dest using net information in NL
void TViewSymbolicNetlist::Dump( ofstream& Dest, TModule* Module, TNetList* NL, bool Descend ) {
  NL = GetNetList( Module, NL );  // replace NL if NL is zero
// first, dump each member part
  for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
    TPart* Part = dynamic_cast<TPart*>( Member );
    if ( Part == 0 )  continue;

// dump part type, name, and reference designator
    Dest << "\n\n" << Part->GetTypeName() << "  "
                   << Part->Name << " (" << Part->GetReferenceDesignator() << "):";
    for ( TCBBase* Member = Part->GetFirstMember(); Member != 0; Member = Member->Next ) {
      TPort* Port = dynamic_cast<TPort*>( Member );
      if ( Port == 0 )  continue;
      for ( int w = Port->GetMaxRange(); w >= Port->GetMinRange(); -- w ) {
        TPortRange PR;
        PR.SetPort( Port );
        PR.SetRange( w, w );

// dump port name
        Dest << "\n  "; Out_PR_NP( Dest, PR );  // output PortRange without path

// dump reference designator and pin(s)
        TPinSpec PinSpec( Port, w );
        Dest << " (";
        Dest << Part->GetReferenceDesignator();
        Dest << '.';
        PinSpec.OutputPins( Dest );
        Dest << ')';

// dump net name
        TNet* Net = NL->PinInNetList( PinSpec );
        if ( Net ) {
          Dest << "  net " << Net;
// dump other connections on net
          if ( !IsBriefNet( Net ) ) {
            for ( unsigned i = 0; i < Net->GetPinSpecCount(); ++ i ) {
              TPinSpec* NPS = Net->GetPinSpec( i );
              if ( !NPS->OwnerIsPart() ) continue;  // as of 14-Aug-2005, TNet's contain all ports, not just ports of parts
              if ( *NPS == PinSpec )     continue;  // only output other PinSpec's
              TPortRange PR;
              PR.SetPort(  NPS->GetPort()  );
              PR.SetRange( NPS->GetIndex(), NPS->GetIndex() );
  // dump the port range
              Dest << "  " << PR;
  // optionally dump the pins
              if ( ShowAllPins ) {
                TPart* Part = dynamic_cast<TPart*>( NPS->GetPort()->Name.Owner );
                Dest << " (";
                if ( Part )  Dest << Part->GetReferenceDesignator();  // should always work
                Dest << '.';
                NPS->OutputPins( Dest, true );   // true = brief mode
                Dest << ')';

              }
            }
          }
        }
        else {
          Dest << " *** no net ***";
        }
      }
    }
  }

// next, if Descend, dump each member module
  if ( Descend ) {
    for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
      TModule* M = dynamic_cast<TModule*>( Member );
      if ( M ) Dump( Dest, M, NL, Descend );
    }
  }
}

// dump each net in netlist
void TViewSymbolicNetlist::DumpNets( ofstream& Dest, TModule* Module, TNetList* NL ) {
  NL = GetNetList( Module, NL );  // replace NL if NL is zero
  for ( unsigned n = 0; n < NL->GetNetCount(); ++ n ) {
    TNet* Net = NL->GetNet( n );
    if ( Net == 0 )  break;   // should never happen

    Dest << "\nnet " << Net << ":";

// dump connections on net
    for ( unsigned i = 0; i < Net->GetPinSpecCount(); ++ i ) {
      TPinSpec* NPS = Net->GetPinSpec( i );
      if ( !NPS->OwnerIsPart() ) continue;  // as of 14-Aug-2005, TNet's contain all ports, not just ports of parts
      TPortRange PR;
      PR.SetPort(  NPS->GetPort()  );
      PR.SetRange( NPS->GetIndex(), NPS->GetIndex() );
  // dump the port range
      Dest << "\n  " << PR;
  // optionally dump the pins
      if ( ShowAllPins ) {
        TPart* Part = dynamic_cast<TPart*>( NPS->GetPort()->Name.Owner );
        Dest << " (";
        if ( Part )  Dest << Part->GetReferenceDesignator();  // should always work
        Dest << '.';
        NPS->OutputPins( Dest, true );   // true = brief mode
        Dest << ')';
      }
    }
  }
}


// perform full dump of each net output with brief format
void TViewSymbolicNetlist::DumpBriefNets( ofstream& Dest, TModule* Module, TNetList* NL ) {
  NL = GetNetList( Module, NL );  // replace NL if NL is zero
  unsigned DumpedNetCount = 0;

  for ( unsigned n = 0; n < NL->GetNetCount(); ++ n ) {
    TNet* Net = NL->GetNet( n );
    if ( Net == 0 )  break;   // should never happen
    if ( !IsBriefNet( Net ) )  continue;
    DumpedNetCount ++;
    Dest << "\nnet " << Net << ":";

// dump connections on net
    for ( unsigned i = 0; i < Net->GetPinSpecCount(); ++ i ) {
      TPinSpec* NPS = Net->GetPinSpec( i );
      if ( !NPS->OwnerIsPart() ) continue;  // as of 14-Aug-2005, TNet's contain all ports, not just ports of parts
      TPortRange PR;
      PR.SetPort(  NPS->GetPort()  );
      PR.SetRange( NPS->GetIndex(), NPS->GetIndex() );
  // dump the port range
      Dest << "\n  " << PR;
  // optionally dump the pins
      if ( ShowAllPins ) {
        TPart* Part = dynamic_cast<TPart*>( NPS->GetPort()->Name.Owner );
        Dest << " (";
        if ( Part )  Dest << Part->GetReferenceDesignator();  // should always work
        Dest << '.';
        NPS->OutputPins( Dest, true );   // true = brief mode
        Dest << ')';
      }
    }
  }
  Dest << "\n\nTotal nets dumped:  " << DumpedNetCount;
  Dest << "\n";
}


// dump parts in Module to Dest
void TViewSymbolicNetlist::DumpParts( ofstream& Dest, TModule* Module, bool Descend ) {
// first, dump each member part
  for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
    TPart* Part = dynamic_cast<TPart*>( Member );
    if ( Part == 0 )  continue;

// dump part type, name, and reference designator
    Dest << "\n"   << Part->GetTypeName() << "  "
                   << Part->Name << " (" << Part->GetReferenceDesignator() << ")";
  }

// next, if Descend, dump each member module
  if ( Descend ) {
    for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
      TModule* M = dynamic_cast<TModule*>( Member );
      if ( M ) DumpParts( Dest, M, Descend );
    }
  }
}



// dump modules in Module to Dest
void TViewSymbolicNetlist::DumpModules( ofstream& Dest, TModule* Module, int Indent, bool Descend ) {
// dump each member module, descend if Descend
  for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
    TModule* MP = dynamic_cast<TModule*>( Member );
    if ( MP == 0 )  continue;

// dump module type and name
    Dest << "\n";
    for ( int i = 0; i < Indent; ++ i ) {
      Dest << ' ';
    }
    Dest << MP->GetTypeName() << ' ';
    Out_D_NP( Dest, MP->Name );  // output name without full hierarchical path

// if Descend, dump each member module
    if ( Descend ) {
        DumpModules( Dest, MP, Indent + 2, Descend );
    }
  }
}


void TViewSymbolicNetlist::DumpHTML( ofstream& Dest, TModule* Module, TNetList* NL ) {
  NL = GetNetList( Module, NL );  // replace NL if NL is zero

  PartPath.Clear();

  Dest << "<html>";
  Dest << "\n<head>";
  Dest << "\n<title>Symbolic Netlist</title>";
  Dest << "\n</head>";
  Dest << "\n<body>";

  Dest << "\n<H1>Symbolic Netlist</H1>";

  Dest << "\n<H2>Module Summary</H2>";
  H_DumpModules( Dest, Module );

  Dest << "\n<BR><H2>Netlist By Part</H2>";
  H_Dump( Dest, Module, NL );
  PartPath.Clear();                                  // no current part for following output --> no brief paths

  Dest << "\n<BR><H2>Non-Enumerated Nets</H2>";
  H_DumpBriefNets( Dest, Module, NL );

  Dest << "\n</body>";
  Dest << "\n</html>";
}


// dump modules in Module to Dest, HTML format
void TViewSymbolicNetlist::H_DumpModules( ofstream& Dest, TModule* Module, int Indent, bool Descend ) {
// dump each member module, descend if Descend
  Dest << "\n<UL>";
  for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
    TModule* MP = dynamic_cast<TModule*>( Member );
    if ( MP == 0 )  continue;

    unsigned ModuleNumber = FindModule( MP );  // get the int associated with the module

// dump module type and name
    Dest << MP->GetTypeName() << ' ';
    Dest << "<A HREF=\"#MOD" << ModuleNumber << "\">";  // e.g., <A HREF="#MOD34">
    Out_D_NP( Dest, MP->Name );  // output name without full hierarchical path
    Dest << "</A>\n";

// if Descend, dump each member module
    if ( Descend ) {
        H_DumpModules( Dest, MP, Indent + 1, Descend );
    }
  }
  Dest << "\n</UL>";
}


// dump Module to Dest using net information in NL, HTML version
void TViewSymbolicNetlist::H_Dump( ofstream& Dest, TModule* Module, TNetList* NL, bool Descend ) {
  NL = GetNetList( Module, NL );  // replace NL if NL is zero

  Dest << "\n<A NAME=\"MOD" << FindModule( Module ) << "\"></A>";  // e.g, <A NAME="MOD34"></A>

// first, dump each member part
  for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
    TPart* Part = dynamic_cast<TPart*>( Member );
    if ( Part == 0 )  continue;

    PartPath.Set( Part );

    Dest << "\n<A NAME=\"PAR" << FindPart( Part ) << "\"></A>";  // e.g, <A NAME="PAR122"></A>

// dump part type, name, and reference designator
    Dest << "\n<BR><B>"  << Part->GetTypeName() << "&nbsp;&nbsp;<font size=+1>";
    Out_ModPar_HTML( Dest, Part, false );
    Dest << " ("  << Part->GetReferenceDesignator() << "):</font></B>";

    Dest << "\n<UL>";
    for ( TCBBase* Member = Part->GetFirstMember(); Member != 0; Member = Member->Next ) {
      TPort* Port = dynamic_cast<TPort*>( Member );
      if ( Port == 0 )  continue;
      for ( int w = Port->GetMaxRange(); w >= Port->GetMinRange(); -- w ) {
        TPortRange PR;
        PR.SetPort( Port );
        PR.SetRange( w, w );

// dump port name
        Dest << "\n"; Out_PR_NP( Dest, PR );  // output PortRange without path

// dump reference designator and pin(s)
        TPinSpec PinSpec( Port, w );
        Dest << " (";
        Dest << Part->GetReferenceDesignator();
        Dest << '.';
        PinSpec.OutputPins( Dest );
        Dest << ')';

// dump net name
        TNet* Net = 0;
        int NetIndex =  PinSpec.GetNetIndex();                           // NetIndex is index in netlist of net containing PinSpec
        if ( NetIndex >= 0 )  Net = NL->GetNet( unsigned( NetIndex ) );  // if net index is assigned
        if ( Net ) {
          bool BriefNet = IsBriefNet( Net );
// optionally dump other connections on net
          if ( ! BriefNet ) {
            Dest << "&nbsp;&nbsp;&nbsp;<B>net " << Net << "</B>";
            for ( unsigned i = 0; i < Net->GetPinSpecCount(); ++ i ) {
              TPinSpec* NPS = Net->GetPinSpec( i );
              if ( !NPS->OwnerIsPart() ) continue;  // as of 14-Aug-2005, TNet's contain all ports, not just ports of parts
              if ( *NPS == PinSpec )  continue;  // only output other PinSpec's
              TPortRange PR;
              PR.SetPort(  NPS->GetPort()  );
              PR.SetRange( NPS->GetIndex(), NPS->GetIndex() );
  // dump the port range
              Dest << "&nbsp;&nbsp;&nbsp;";
              Out_PR_HTML( Dest, PR );
  // optionally dump the pins
              if ( ShowAllPins ) {
                TPart* Part = dynamic_cast<TPart*>( NPS->GetPort()->Name.Owner );
                Dest << " (";
                if ( Part )  Dest << Part->GetReferenceDesignator();  // should always work
                Dest << '.';
                NPS->OutputPins( Dest, true );   // true = brief mode
                Dest << ')';
              }
            }
          }
          else {    // brief nets:  link to separate dump of brief net
            Dest << "&nbsp;&nbsp;&nbsp;<B>net</B> ";
            Dest << "<A HREF=\"#NET" << NetIndex << "\">";  // e.g., <A HREF="#NET88">
            Dest << Net;
            Dest << "</A>\n";
          }
        }
        else {
          Dest << "<B> *** no net ***</B>";
        }
        Dest << "<BR>";
      }
    }
    Dest << "\n</UL>";
  }

// next, if Descend, dump each member module
  if ( Descend ) {
    for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
      TModule* M = dynamic_cast<TModule*>( Member );
      if ( M ) H_Dump( Dest, M, NL, Descend );
    }
  }
}


// dump each net in netlist, html version
void TViewSymbolicNetlist::H_DumpNets( ofstream& Dest, TModule* Module, TNetList* NL ) {
  NL = GetNetList( Module, NL );  // replace NL if NL is zero
  for ( unsigned n = 0; n < NL->GetNetCount(); ++ n ) {
    TNet* Net = NL->GetNet( n );
    if ( Net == 0 )  break;   // should never happen

    Dest << "\nnet " << Net << ":";

// dump connections on net
    for ( unsigned i = 0; i < Net->GetPinSpecCount(); ++ i ) {
      TPinSpec* NPS = Net->GetPinSpec( i );
      if ( !NPS->OwnerIsPart() ) continue;  // as of 14-Aug-2005, TNet's contain all ports, not just ports of parts
      TPortRange PR;
      PR.SetPort(  NPS->GetPort()  );
      PR.SetRange( NPS->GetIndex(), NPS->GetIndex() );
  // dump the port range
      Dest << "\n  " << PR;
  // optionally dump the pins
      if ( ShowAllPins ) {
        TPart* Part = dynamic_cast<TPart*>( NPS->GetPort()->Name.Owner );
        Dest << " (";
        if ( Part )  Dest << Part->GetReferenceDesignator();  // should always work
        Dest << '.';
        NPS->OutputPins( Dest, true );   // true = brief mode
        Dest << ')';
      }
    }
  }
}


// perform full dump of each net output with brief format, HTML version
void TViewSymbolicNetlist::H_DumpBriefNets( ofstream& Dest, TModule* Module, TNetList* NL ) {
  NL = GetNetList( Module, NL );  // replace NL if NL is zero

  unsigned DumpedNetCount = 0;
  for ( unsigned n = 0; n < NL->GetNetCount(); ++ n ) {
    TNet* Net = NL->GetNet( n );
    if ( Net == 0 )  break;   // should never happen
    if ( ! IsBriefNet( Net ) ) continue;
    DumpedNetCount ++;

    Dest << "\n<A NAME=\"NET" << n << "\"></A>";  // e.g, <A NAME="NET99"></A>
    Dest << "\n<BR><B>net " << Net << ":</B>";
// dump connections on net
    Dest << "\n<UL>";
    for ( unsigned i = 0; i < Net->GetPinSpecCount(); ++ i ) {
      TPinSpec* NPS = Net->GetPinSpec( i );
      if ( !NPS->OwnerIsPart() ) continue;  // as of 14-Aug-2005, TNet's contain all ports, not just ports of parts
      TPortRange PR;
      PR.SetPort(  NPS->GetPort()  );
      PR.SetRange( NPS->GetIndex(), NPS->GetIndex() );
  // dump the port range
      Out_PR_HTML( Dest, PR );
  // optionally dump the pins
      if ( ShowAllPins ) {
        TPart* Part = dynamic_cast<TPart*>( NPS->GetPort()->Name.Owner );
        Dest << " (";
        if ( Part )  Dest << Part->GetReferenceDesignator();  // should always work
        Dest << '.';
        NPS->OutputPins( Dest, true );   // true = brief mode
        Dest << ')';
      }
      Dest << "<BR>\n";
    }
    Dest << "\n</UL>";
  }
  Dest << "\n<BR><BR>Total nets dumped:  " << DumpedNetCount;
  Dest << "\n<BR>";
}



// dump parts in Module to Dest, HTML version
void TViewSymbolicNetlist::H_DumpParts( ofstream& Dest, TModule* Module, bool Descend ) {
// first, dump each member part
  for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
    TPart* Part = dynamic_cast<TPart*>( Member );
    if ( Part == 0 )  continue;

// dump part type, name, and reference designator
    Dest << "\n"   << Part->GetTypeName() << "  "
                   << Part->Name << " (" << Part->GetReferenceDesignator() << ")";
  }

// next, if Descend, dump each member module
  if ( Descend ) {
    for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {
      TModule* M = dynamic_cast<TModule*>( Member );
      if ( M ) DumpParts( Dest, M, Descend );
    }
  }
}



ostream& TViewSymbolicNetlist::Out_ModPar_HTML( ostream& Dest, TModPar* ModPar, bool PartLink ) {
  if ( ModPar == 0 )  return Dest;

// Output part name, including full path with links to modules
// Output link to part only if PartLink = true

  TPath     Path      = TPath( ModPar );
  size_t    LastIndex = Path.ModPars.size() - 1;    // index of ModPar within Path

  for ( size_t i = 1; i <= LastIndex; ++ i ) {      // for each module/part in path, excluding root
    TModPar* Ancestor    = Path.ModPars[ i ];
    bool     Last        = ( i == LastIndex );
    bool     OutputBrief = BriefPathChar && PartLink && ( i < PartPath.ModPars.size() );  // ok to replace ancestor name with BriefPathChar if name corresponds with ancestor name in current part's path?
    if ( OutputBrief )  OutputBrief = Ancestor->Name.DesignatorEquals( PartPath.ModPars[ i ]->Name );

    if ( Last ) {          // if last ModPar in Path, then output link to part (Ancestor should be same as ModPar)
      if ( PartLink ) {
        TPart* PP = dynamic_cast<TPart*>( Ancestor );
        if ( PP ) {
          unsigned PartNumber = FindPart( PP );             // get the int associated with the Part
          Dest << "<A HREF=\"#PAR" << PartNumber << "\">";  // e.g., <A HREF="#PAR34">
        }
        else PartLink = false;
      }
    }
    else {                 // else output link to ancestor module
      TModule* MP = dynamic_cast<TModule*>( Ancestor );
      if ( MP ) {                                           // MP should never be zero
        unsigned ModuleNumber = FindModule( MP );           // get the int associated with the module
        Dest << "<A HREF=\"#MOD" << ModuleNumber << "\">";  // e.g., <A HREF="#MOD34">
      }
    }

    // output the link's text
    if ( OutputBrief )            Dest << BriefPathChar;      // output BriefPathChar instead of name
    else                Out_D_NP( Dest, Ancestor->Name );     // output name without full path

    if ( ! Last )  Dest  << "</A>.";
  }

  if ( PartLink ) {
    Dest << "</A>";
  }

  return Dest;
}


ostream& TViewSymbolicNetlist::Out_PR_HTML( ostream& Dest, TPortRange& PR, bool PartLink ) {

// Output full port range, including full path with links to modules and part (if PartLink = true)

// first the full path to the part
  Out_ModPar_HTML( Dest, PR.GetPort()->GetOwner(), PartLink );

  Dest << '.';

// next, output the port range
  Out_PR_NP( Dest, PR );

  return Dest;
}

