/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_views.h:  symbolic netlist generator

#ifndef _cb_viewsH_
#define _cb_viewsH_

#include "cb_base.h"

class TViewSymbolicNetlist {
private:
  TPath            PartPath;       // path of current part, e.g., part being output by Out_ModPar_HTML
  int              BriefNetLimit;  // nets with this many PinSpec's or more will be output with brief format
  char             BriefPathChar;  // nonzero --> enable brief path notation:
                                   //   applies only to paths in connections listed within a part's connection list
                                   //   a module name will be replaced with BriefPathChar if the name matches the name of the corresponding module in the path of the part whose connections are being listed
  vector<char*>    BriefNets;

// "Find" in function names below is an anachronism -- vectors of pointers were formerly used to assign unique numbers to each module and part

  unsigned FindModule( TModule* MP ) {    // return unique module number
    return MP->SerialNumber;
  }

  unsigned FindPart( TPart* MP ) {        // return unique part number
    return MP->SerialNumber;
  }

  TNetList* GetNetList( TModule* aModule, TNetList* aNetList ) {   // function added to allow user to specify TRootModule instead of TModule + TNetList
    if ( aNetList ) return aNetList;
    else {                                                         // if no netlist specified, try to use netlist in root module
      TRootModule* RM = dynamic_cast<TRootModule*>( aModule );     // this code is not foolproof, e.g.,
      if ( ! RM )  BEGERR << "TViewSymbolicNetlist::GetNetList(): no NetList and/or RootModule specified" << ENDERR;
      return &RM->NetList;
    }
  }

  ostream& Out_ModPar_HTML( ostream& Dest, TModPar*    ModPar, bool PartLink = true );
  ostream& Out_PR_HTML(     ostream& Dest, TPortRange&     PR, bool PartLink = true );

  // HTML versions of dump functions
  void H_Dump(          ofstream& Dest, TModule* Module, TNetList* NL, bool Descend = true );
  void H_DumpNets(      ofstream& Dest, TModule* Module, TNetList* NL );
  void H_DumpBriefNets( ofstream& Dest, TModule* Module, TNetList* NL );
  void H_DumpParts(     ofstream& Dest, TModule* Module, bool Descend = true );
  void H_DumpModules(   ofstream& Dest, TModule* Module, int Indent = 0, bool Descend = true );

public:
  bool ShowAllPins;
  void SetBriefPathChar( char aBriefPathChar = 0 ) { BriefPathChar = aBriefPathChar; }
  void SetBriefNetLimit( int  aBriefNetLimit = 4 ) { BriefNetLimit = aBriefNetLimit; }
  void ClearBriefNets() {
    BriefNets.clear();
  }

  bool IsBriefNet( TNet* aNet ) {             // return true if net should be output with brief format -- typically not called by user, but public in case it is of use some day
    if ( aNet == 0 )                                            return true;
    if ( int( aNet->GetPartPinSpecCount() ) >= BriefNetLimit )  return true;
    for ( unsigned i = 0; i < BriefNets.size(); ++ i ) {
      if ( strcmp( BriefNets[ i ], aNet->GetString() ) == 0 )   return true;
    }
    return false;
  }

  // add net(s) to list of nets to be output in brief format
  void AddBriefNet( char* Net0,
                    char* Net1 = 0, char* Net2 = 0, char* Net3 = 0,
                    char* Net4 = 0, char* Net5 = 0, char* Net6 = 0,
                    char* Net7 = 0, char* Net8 = 0, char* Net9 = 0 ) {
                BriefNets.push_back( Net0 );
    if ( Net1 ) BriefNets.push_back( Net1 );
    if ( Net2 ) BriefNets.push_back( Net2 );
    if ( Net3 ) BriefNets.push_back( Net3 );
    if ( Net4 ) BriefNets.push_back( Net4 );
    if ( Net5 ) BriefNets.push_back( Net5 );
    if ( Net6 ) BriefNets.push_back( Net6 );
    if ( Net7 ) BriefNets.push_back( Net7 );
    if ( Net8 ) BriefNets.push_back( Net8 );
    if ( Net9 ) BriefNets.push_back( Net9 );
  }
  // dump Module to Dest using net information in NL
  void Dump(          ofstream& Dest, TModule* Module, TNetList* NL = 0, bool Descend = true );
  void DumpNets(      ofstream& Dest, TModule* Module, TNetList* NL = 0 );
  void DumpBriefNets( ofstream& Dest, TModule* Module, TNetList* NL = 0 );
  void DumpParts(     ofstream& Dest, TModule* Module, bool Descend = true );
  void DumpModules(   ofstream& Dest, TModule* Module, int Indent = 0, bool Descend = true );
  void DumpHTML(      ofstream& Dest, TModule* Module, TNetList* NL = 0 );
  TViewSymbolicNetlist() {
    SetBriefPathChar();       // set defaults
    SetBriefNetLimit();
    ShowAllPins    = false;
  }
};

#endif

