/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_xconstraints.cpp:  location constraints file generator for Xilinx FPGA design tools (Xilinx is a registered trademark of Xilinx, Inc.)

#include "cb_xconstraints.h"

bool operator<( TXConstraintSpec A, TXConstraintSpec B ) { return  A.Name < B.Name; }


void TXConstraints::ProcessPort( TPort* P ) {         // process port of target module
  for ( int i = P->GetMinRange(); i <= P->GetMaxRange(); i++ ) {
    PS = TPinSpec( P, i );     // target module's current PinSpec
    TNet* Net = ModuleNetList->PinInNetList( PS );
    if ( Net ) {
      Trace( Net );   // go check all the PinSpecs in the net
    }
  }
}


void TXConstraints::Trace( TNet* Net, bool Init ) {     // trace from specified net to port of target part
  if ( Init ) {
    TraceNets.clear();
    TraceParts = 0;
  }
  if ( ! Net )  return;
  for ( unsigned i = 0; i <  TraceNets.size(); i++ ) {
    if ( Net == TraceNets[ i ] )  return;                     // if net was already traced through then do not
  }
  
  for ( unsigned j = 0; j < Net->GetPinSpecCount(); j++ ) {   // for each PinSpec in Net
    TPinSpec* NetPS     = Net->GetPinSpec( j );
    TPort*    Port      = NetPS->GetPort();
    TPart*    TracePart = dynamic_cast<TPart*>( Port->GetOwner() );
    if ( ! TracePart                                 )  continue;    // ignore ports that are not owned by part
    if ( Port->MultiPinNode()                        )  continue;    // ignore multi-pin nodes
    if ( TracePart == Part ) { SaveSpec( NetPS );       continue; }
//  if ( Net->GetPinSpecCount() >  TraceMaxNetSize   )  continue;    // test this below instead  // test this after Part/TracePart check --> IgnorePorts can apply to large nets on part pins, e.g., pins tied high or low
    if ( TraceParts             >= MaxThroughParts   )  continue;
    if ( !WithinTarget( TracePart )                  )  continue;    // if part is not within target module
    if ( TracePart->GetPackage().GetPinCount() != 2  )  continue;    // if not two-terminal part

    TPort* P;  int Index;
    for ( TCBBase* Member = TracePart->GetFirstMember(); Member != 0; Member = Member->Next ) {    // find the part's other pin
      P = dynamic_cast<TPort*>( Member );
      if ( ! P )  continue;                       // should never happen
      Index = P->GetMinRange();                   // possible index for PinSpec of OtherPin
      if ( P == Port ) {
        if ( P->GetPinCount() == 1 )  continue;   // other pin must be in another port
        if ( Index == NetPS->GetIndex() ) Index = P->GetMaxRange();
      }
      break;
    }
    if ( ! P )  continue;            // should never happen
    TPinSpec OtherPin( P, Index );
    TNet* NextNet = ModuleNetList->PinInNetList( OtherPin );
    if ( NextNet->GetPinSpecCount()  >  TraceMaxNetSize )  continue;    // never trace through a part to a large net

    TraceNets.push_back( Net );      // save current net in list to avoid endless loop
    TraceParts ++;
    Trace( NextNet, false );         // recurse, do not init
    TraceParts --;
  }
}

void TXConstraints::SaveSpec( TPinSpec* PartPS ) {     // save current PS and PartPS in  ConstraintSpecs
  TPort* P = PS.GetPort();     // from current target module PS
  int    i = PS.GetIndex();

  TXConstraintSpec  Spec;
  Spec.Name         = D_NP( P->Name );
  // if both LeftRange and RightRange are zero, then P has no specified range (single wire)
  if ( P->GetLeftRange() || P->GetRightRange() )  Spec.Name += CSTRING( "<" << i << ">" );
  Spec.PartPinSpec  =  *PartPS;
  TPort* PP         =  PartPS->GetPort();

  bool Traced       =  ( TraceParts <= TraceThroughParts  )  &&  UseModulePorts;                            // only keep   if trace depth for traces  was not exceeded
  bool IgnorePort   =  ( TraceParts <= IgnoreThroughParts )  &&  IsInIgnoreList( P,  IgnorePorts     );     // only ignore if trace depth for ignores was not exceeded
  Spec.Ignore       =  IgnorePort                            ||  IsInIgnoreList( PP, IgnorePartPorts );     // set flag if module's port is in ignore list --> subsequent code will ignore this part PinSpec

  if ( Traced || Spec.Ignore ) {
    PushSpec( Spec, true );            // push Spec, override existing spec if its Ignore member is true
  }
}


bool TXConstraints::PushSpec( TXConstraintSpec& Spec, bool OverrideIgnore ) {     // push Spec onto ConstraintSpecs but only if its PartPinSpec is not already in ConstraintSpecs
  for ( vector<TXConstraintSpec>::iterator I = ConstraintSpecs.begin(); I != ConstraintSpecs.end(); I++ ) {
    if ( Spec.PartPinSpec ==  I->PartPinSpec ) {
      if ( I->Ignore && OverrideIgnore ) {     // override existing ignored PinSpec, e.g., pin with pulldown to GND
        ConstraintSpecs.erase( I );            // remove...
        break;                                 // ...add at back --> order is not determined by overridden Spec
      }
      return false;
    }
  }
  ConstraintSpecs.push_back( Spec );
  return true;
}


void TXConstraints::ProcessBundle( TBundle* C ) {       // process bundle of target module
  BundlePrefix += D_NP( C->Name ) + " ";                // accumulate bundle instance names, with terminal space
//cout << "\n" << BundlePrefix;
  for ( TCBBase* Member = C->GetFirstMember(); Member != 0; Member = Member->Next ) {  // for each member of the bundle, it is either a port or an embedded bundle
    TPort*     P     =  dynamic_cast<TPort*>(   Member );
    TBundle*   B     =  dynamic_cast<TBundle*>( Member );
    if (       P )  ProcessPort(   P );
    else if (  B )  ProcessBundle( B );
  }
}


void TXConstraints::AddIgnorePort( char* Port0,
                  char* Port1, char* Port2, char* Port3,
                  char* Port4, char* Port5, char* Port6,
                  char* Port7, char* Port8, char* Port9 ) {
               IgnorePorts.push_back( Port0 );
  if ( Port1 ) IgnorePorts.push_back( Port1 );
  if ( Port2 ) IgnorePorts.push_back( Port2 );
  if ( Port3 ) IgnorePorts.push_back( Port3 );
  if ( Port4 ) IgnorePorts.push_back( Port4 );
  if ( Port5 ) IgnorePorts.push_back( Port5 );
  if ( Port6 ) IgnorePorts.push_back( Port6 );
  if ( Port7 ) IgnorePorts.push_back( Port7 );
  if ( Port8 ) IgnorePorts.push_back( Port8 );
  if ( Port9 ) IgnorePorts.push_back( Port9 );
}


void TXConstraints::AddIgnorePartPort( char* Port0,
                  char* Port1, char* Port2, char* Port3,
                  char* Port4, char* Port5, char* Port6,
                  char* Port7, char* Port8, char* Port9 ) {
               IgnorePartPorts.push_back( Port0 );
  if ( Port1 ) IgnorePartPorts.push_back( Port1 );
  if ( Port2 ) IgnorePartPorts.push_back( Port2 );
  if ( Port3 ) IgnorePartPorts.push_back( Port3 );
  if ( Port4 ) IgnorePartPorts.push_back( Port4 );
  if ( Port5 ) IgnorePartPorts.push_back( Port5 );
  if ( Port6 ) IgnorePartPorts.push_back( Port6 );
  if ( Port7 ) IgnorePartPorts.push_back( Port7 );
  if ( Port8 ) IgnorePartPorts.push_back( Port8 );
  if ( Port9 ) IgnorePartPorts.push_back( Port9 );
}


void TXConstraints::AddIgnoreNet( char* Port0,
                  char* Port1, char* Port2, char* Port3,
                  char* Port4, char* Port5, char* Port6,
                  char* Port7, char* Port8, char* Port9 ) {
               IgnoreNets.push_back( Port0 );
  if ( Port1 ) IgnoreNets.push_back( Port1 );
  if ( Port2 ) IgnoreNets.push_back( Port2 );
  if ( Port3 ) IgnoreNets.push_back( Port3 );
  if ( Port4 ) IgnoreNets.push_back( Port4 );
  if ( Port5 ) IgnoreNets.push_back( Port5 );
  if ( Port6 ) IgnoreNets.push_back( Port6 );
  if ( Port7 ) IgnoreNets.push_back( Port7 );
  if ( Port8 ) IgnoreNets.push_back( Port8 );
  if ( Port9 ) IgnoreNets.push_back( Port9 );
}

void  TXConstraints::GenerateConstraintsFile( char*  FileName,               // destination file name
                                                     TModule*  aModule,      // target module, 0 --> use names from hierarchy's net list
                                                     TPart*    aPart ) {     // target part

  if ( aPart == 0       )  return;    // should never happen
  if ( aPart->Root == 0 )  return;    // should never happen

  Module          = aModule;
  Part            = aPart;
  DesignNetList   = &Part->Root->NetList;

  MaxThroughParts = TraceThroughParts;
  if ( IgnoreThroughParts > MaxThroughParts )  MaxThroughParts = IgnoreThroughParts;

  ConstraintsFile.open( FileName, ios_base::out );
  if ( ! ConstraintsFile ) {
    BEGERR << "TXConstraints::GenerateConstraintsFile() could not open " << FileName << " for writing" << ENDERR;
  }

  ConstraintSpecs.clear();

  ProcessTargetModule();       // if user specified a target module, trace its PinSpecs to the target part
  ProcessDesignNetList();      // use design netlist to assign names to target part's pins that have no name (if enabled via UseDesignNetNames)
  PostProcess();               // perform all remaining processing, including output to destination

  ConstraintsFile.close();
}


void  TXConstraints::ProcessTargetModule(){
  if ( Module == 0 )  return;                // if user did not specify a target module

  TNetList NL;                               // create a netlist just for the target module
  ModuleNetList      = &NL;                  // for use by ProcessPort(), etc.
  Module->FillNetList( &NL );
  NL.Cleanup();

  for ( TCBBase* Member = Module->GetFirstMember(); Member != 0; Member = Member->Next ) {    // Go through all the ports in the module
    BundlePrefix = "";
    TPort*    P  =  dynamic_cast<TPort*>(   Member );
    TBundle*  B  =  dynamic_cast<TBundle*>( Member );
    if (      P ) ProcessPort( P );         // if member is a port, find the pins it is connected to
    else if ( B ) {
      BundlePrefix = "";
      ProcessBundle( B );                   // if member is a bundle, descend into it
    }
  }
  DesignNetList->SetNetIndexes();           // restore net indexes that were changed when ModuleNetList was filled
};


void  TXConstraints::ProcessDesignNetList(){
  if ( DesignNetList == 0  )  return;        // should never happen
  if ( ! UseDesignNetNames )  return;        // if user does not want design netlist to be used to generate LOC constraints (except for tracing through parts)
  for ( TCBBase* Member = Part->GetFirstMember(); Member != 0; Member = Member->Next ) {    // Go through all the ports in the part
    TPort*  P  =  dynamic_cast<TPort*>( Member );   // all members should be ports, but check anyway
    if ( ! P )                                   continue;
    if ( IsInIgnoreList( P, IgnorePartPorts ) )  continue;
    if ( P->MultiPinNode() )                     continue;      // ignore multi-pin nodes

    for ( int i = P->GetMinRange(); i <= P->GetMaxRange(); i++ ) {  // for each node in the port
      bool Found = false;
      TPinSpec PS( P, i );     // target part's current PinSpec
      for ( unsigned s = 0; s < ConstraintSpecs.size(); ++ s ) {
        if ( PS == ConstraintSpecs[ s ].PartPinSpec ) { Found = true; break; }
      }
      if ( Found )  continue;  // ignore all PinSpec's that already have a ConstraintSpec

      TNet* DesignNet = DesignNetList->PinInNetList( PS );
      if ( DesignNet == 0 )  continue;
      if ( DesignNet->DesignatorEquals( NoConnect ) ) {   // if no-connect net
        if ( IgnoreNoConnect )  continue;
      }
      else if ( IgnoreGlobal && ( DesignNet->GetString()[ 0 ] == GlobalChar ) ) {   // if global net but not no-connect
        continue;
      }

      if ( IsInIgnoreList( DesignNet, IgnoreNets ) )  continue;

      TXConstraintSpec  Spec;
      Spec.Name         = NETN( DesignNet );  // net name in form: name[__extension][<number>]
      Spec.PartPinSpec  = PS;
      PushSpec( Spec );
//    cout  << "\nnet  " << Spec.Name;
    }
  }
}


void TXConstraints::PostProcess() {          // perform all remaining processing based on contents of ConstraintSpecs

// handle ForceUpperCase and RemoveGlobalChar
  for ( unsigned i = 0; i < ConstraintSpecs.size(); ++ i ) {
    TXConstraintSpec* Spec = & ConstraintSpecs[ i ];
    if ( Spec->Ignore )  continue;                      // unnecessary -- for consistency with other loops
    if ( ForceUpperCase ) {
      ToUpper( Spec->Name       );
      ToUpper( Spec->NamePrefix );
    }
    if ( RemoveGlobalChar ) {      // remove GlobalChar from beginning of net name if enabled
      if ( Spec->Name.size() > 0 ) {
        if ( Spec->Name[ 0 ] == GlobalChar )  Spec->Name.erase( 0, 1 );   // erase first character
      }
    }
  }


  for ( unsigned i = 0; i < ConstraintSpecs.size(); ++ i ) {
    TXConstraintSpec* Spec = & ConstraintSpecs[ i ];
    if ( Spec->Ignore )  continue;
    ConstraintsFile << "NET \"" << Spec->Name;           // net name including <index> if applicable
    ConstraintsFile << "\" LOC = \"" << PinPrefix << Spec->PartPinSpec.GetPin( 0 ) << "\";";  // 0 is ignored because not multi-pin node
    if ( CommentPinName ) {
      TPort* P = Spec->PartPinSpec.GetPort();
      ConstraintsFile << "    # ";
      Out_PR_NP( ConstraintsFile, (*P)( Spec->PartPinSpec.GetIndex() ) );
    }
    ConstraintsFile << "\n";
  }



  // !!! generate unique -- not currently supported
  // !!! align output    -- not currently supported
}

