/*                                                                               *\
    This file is part of the CBOLD(tm) framework for electronic design capture.
                         Copyright (C) 2007 Coolquest, Inc.

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License along
      with this program; if not, write to the Free Software Foundation, Inc.,
      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For more about the CBOLD framework and Coolquest, Inc., see www.cbold.com.
\*                                                                               */

// file cb_xconstraints.h:  location constraints file generator for Xilinx FPGA design tools (Xilinx is a registered trademark of Xilinx, Inc.)

#ifndef _cb_xconstraintsH_
#define _cb_xconstraintsH_

#include "cb_base.h"

// TXConstraints is a class that generates constraint files compatible with Xilinx UCF format.
// When GenerateConstraintsFile() is called, it is passed a target module (optional) and a target part.
// The generated constraint file contains pin location constraints for the target part.
// For each of the target module's PinSpec's (i.e., node of a port) GenerateConstraintFiles() traces the connection from the PinSpec to the target part.
// If the trace is successful, the pin name in the constraint file is derived from the target module's port name.
// If the trace fails for a particular PinSpec, the following is attempted if enabled:
// The connection is traced through certain parts, e.g., through a resistor that is connected between the PinSpec and the target part.
// If the trace still fails, the PinSpec is ignored, i.e., the PinSpec has no effect on the part's constraint file.

// Once all of the target module's PinSpec's are traced, GenerateConstraintsFile() checks that each of the target part's pins has a location constraint in the constraint file.
// If not, the following is attempted if enabled:
//   A location constraint is derived from the name (in the design's netlist) of the net attached to the pin (if any).
// Multi-pin nodes of the target part (e.g., power supply pins) never contribute to the constraints file.

// If enabled to do so, GenerateConstraintsFile() will modify pin names to eliminate duplicates, force pin names to uppercase, etc.
// If multiple PinSpecs of the target module connect to the same pin on the part, then multiple LOC constraints will be created.

// In general, the target module should be the highest-level wrapper module for the target part.
// For example, an FPGA is typically found in a hierarchy such as:
//   CM_BusMasterFpga       // target module  application-specific wrapper
//     CM_StdFpga           //                generic wrapper -- member of application-specif wrapper
//       CP_XC2S150_FG256   // target part    member of generic wrapper

/* The general algorithm for GenerateConstraintsFile() is:
  if ( aModule != 0 ) {
    for each port of Module {
      trace each node of port to a PinSpec on Part (including trace through parts if enabled)
      if successful {
        if ( Module's port is in IgnorePorts list ) mark the PinSpec as ignored
        else if ( UseModulePorts )                  mark the PinSpec as included in the output file
      }
    }
  }
  if ( UseDesignNetNames ) {
    for each PinSpec on Part {
      if marked ignored                                             then continue;
      if marked included                                            then continue;
      if PinSpec is connected to Part port in IgnorePartPorts list  then continue;
      if PinSpec is not connected to a net                          then continue;
      if PinSpec is connected to net in IgnoreNets list             then continue;
      derive a name for the PinSpec from the net it is attached to and mark the PinSpec as included in the output file
    }
  }
*/

//  In addition to the above algorithm:
//    no multi-pin node of Part will appear in the output file
//    flags and other members listed below will apply

// PinPrefix                   the string to be prefixed on pin names in the UCF file, e.g., some parts in QFP's require a "P" before the pin number
//                             max number of parts to trace through between port on the target module and port on the target part, ignored if Module is 0 or if DesignNetList is 0:
// TraceThroughParts           > 0  --> allow connections to be traced through at most TraceThroughParts two-terminal parts
// IgnoreThroughParts          > 0  --> same as above, but only for purposes of ignoring ports on the target module, e.g., to ignore part ports tied to GND use 0, to ignore part ports pulled to GND through a single resistor, use 1
// TraceMaxNetSize             max size of a net that will be traced through (in PinSpec's) -- e.g., this value prevents tracing through power and ground nets
// ForceUpperCase              true --> convert pin names in the UCF file to upper case
// GenerateUnique              true --> generate unique pin names when duplicate names are encountered, e.g., when the same bundle is instantiated twice in the target module
// CommentPinName              true --> adds comments in the UCF file that contain the names of pins as they appear in the target TPart
// AlignOutput                 true --> aligns columns in output if practical
// IgnoreGlobal                true --> do not derive a location constraint from a net name if the net is a global net (except the no-connect net)
// IgnoreNoConnect             true --> do not derive a location constraint from a net name if the net is no-connect net
// RemoveGlobalChar            true --> remove the first character from global net names, e.g., to create a name legal as a pin name
// UseModulePorts              true --> use target module's ports to form constraint file, otherwise just use target module to specify ignored ports
// UseDesignNetName            true --> use net name from the design netlist if a suitable port name cannot be used (e.g., if a part's pin cannot be traced to a port of the target module)
// AddIgnorePort()             adds port(s) to the list of ports of the target module to be ignored
// AddIgnorePartPort()         adds port(s) to the list of ports of the target part   to be ignored
// AddIgnoreNet()              adds net(s)  to the list of nets to be ignored -- applies only to nets used as a result of UseDesignNetList being true
// ClearIgnore...()            clears the corresponding ignore list
// GenerateConstraintsFile()   generates the constraint file based on the values/calls above as well as its arguments:
//   char*     FileName                 // destination file name
//   TModule*  aModule                  // target module, 0 --> use names from design netlist only
//   TPart*    aPart                    // target part

// notes on AddIgnore...() calls:
//   net and port names consist of several parts:
//      string[extension][number]
//        the extension is typically used to distinguish net names that would otherwise be identical, e.g., DONE, DONE__1, DONE__2
//        the number is typically used as an index in a port or a set of related nets, e.g., D(0), D(1), ..., D(31)
//   strings in the ignore list are compared only to the fundamental string in the net or port name -- the extension and number are not considered
//   a terminal wildcard (*) is allowed in arguments of AddIgnore...(), e.g., "VCC*" matches "VCC", "VCC3_3", etc.



class TXConstraintSpec {
public:
  string    Name;          // name to be used for pins -- includes <index> if applicable
  string    NamePrefix;    // name(s) of bundle(s) that contain target module PinSpec's port, separated with spaces -- used only if Name is not unique
  TPinSpec  PartPinSpec;   // target part's pin to which constraint applies
  bool      Ignore;        // true --> part pin was determined to be one that should be ignored
  bool      Comment;       // true --> output comment instead of LOC constraint, e.g., due to multiple
};
bool operator<( TXConstraintSpec A, TXConstraintSpec B );


class TXConstraints {
private:
  enum  { WildChar = '*' };
  vector<char*>             IgnorePorts;
  vector<char*>             IgnorePartPorts;
  vector<char*>             IgnoreNets;
  vector<TXConstraintSpec>  ConstraintSpecs;
  vector<TNet*>             TraceNets;          // used to avoid endless loop during Trace()
  unsigned                  TraceParts;         // used to stop trace when >= TraceThroughParts
  unsigned                  MaxThroughParts;    // max of TraceThroughParts and IgnoreThroughParts
  ofstream                  ConstraintsFile;

  TModule*       Module;           // copies of GenerateConstraintsFile() arguments
  TPart*         Part;
  TNetList*      DesignNetList;

  TNetList*      ModuleNetList;    // netlist of Module only -- the netlist is created and filled by ProcessTargetModule()
  string         BundlePrefix;     // name(s) of bundle(s) currently being processed, separated with spaces
  TPinSpec       PS;               // target module's current PinSpec during GenerateConstraintsFile()


  bool IsInIgnoreList( const char* aS, vector<char*>& List ) {
    for ( unsigned i = 0; i < List.size(); i++ ) {     // for each entry in list
      const char*  L = List[ i ];
      const char*  S = aS;
      bool   Match = false;
      for ( int j = 0; j < 256; ++ j ) {               // while( 1 )  >>> assume no one has net names or port names longer than 256 chars
        char cl = *L++;  char cs = *S++;
        Match = ( cs == cl ) || ( cl == WildChar );
        if ( ! Match                           )  break;
        if ( ( cl == 0 ) || ( cl == WildChar ) )  break;
      }
      if ( Match )  return true;
    }
    return false;
  }

  bool IsInIgnoreList( TPort* P, vector<char*>& List ) {
    return  IsInIgnoreList( P->Name.GetString(), List );   // just check string, ignore number
  }

  bool IsInIgnoreList( string& S, vector<char*>& List ) {
    return  IsInIgnoreList( S.c_str(), List );
  }

  bool IsInIgnoreList( TNet* N, vector<char*>& List ) {
    return  IsInIgnoreList( N->GetString(), List );        // just check string, ignore extension and number
  }

  bool WithinTarget( TCBBase* CBB ) {     // return true if CBB is owned by target module or one of its members
    while ( CBB != 0 ) {
      if ( CBB->GetOwner() == Module )  return true;
      CBB = CBB->GetOwner();
    }
    return false;
  }

  string D_NP( TDesignator D ) {   // convert designator to string with no path
    ostringstream Temp;            // = ostringstream().flush();
    Out_D_NP( Temp, D );           // use standard output function to output designator to Temp
    return  Temp.str();
  }

  string NETN( TNet* NN, bool UseExtension = false ) {        // return string containing net name in form name[__extension][<number>]   (__extension is omitted by default)
    ostringstream Temp;            // = ostringstream().flush();
    char* B =  NN->GetBase();      // code derived from ostream& Out_D_NP( ostream& Dest, TDesignator D )
    if ( B )                        Temp <<  B;                   // base is unlikely for a net name
                                    Temp <<  NN->GetString();
    if ( UseExtension ) {
      if ( NN->GetExtension() != 0 )  Temp << "__" << NN->GetExtension();
    }
    int  N = NN->GetNumber();
    if ( N != -1 )                  Temp << "<" << N << ">";      // e.g., if bus net
    return  Temp.str();
  }

  void ToUpper( string& s ) {
    string UpString = "";
    const char* c = s.c_str();
    while ( *c != 0 )  UpString += toupper( *c++ );
  }

  void ProcessTargetModule();
  void ProcessDesignNetList();
  void PostProcess();
  void ProcessPort(   TPort*   P );
  void ProcessBundle( TBundle* C );
  void Trace( TNet* Net, bool Init = true );                                 // trace from specified net to port of target part
  void SaveSpec( TPinSpec* PartPS );                                         // save current PS and PartPS in  ConstraintSpecs
  bool PushSpec( TXConstraintSpec& Spec, bool OverrideIgnore = false );      // push Spec onto ConstraintSpecs but only if its PartPinSpec is not already in ConstraintSpecs, return true if it is pushed


public:
                                   // the user should set these before calling GenerateConstraintsFile()
  char*          PinPrefix;        // e.g., some FPGA's in QFP's require a "P" before the pin number
  unsigned       TraceThroughParts;
  unsigned       IgnoreThroughParts;
  unsigned       TraceMaxNetSize;  // the max size of a net that will be traced through (in PinSpec's) -- e.g., this value prevents tracing through power and ground nets 
  bool           ForceUpperCase;
  bool           GenerateUnique;   // !!! not currently supported
  bool           CommentPinName;
  bool           AlignOutput;      // !!! not currently supported
  bool           IgnoreGlobal;
  bool           IgnoreNoConnect;
  bool           RemoveGlobalChar;
  bool           UseModulePorts;
  bool           UseDesignNetNames;

  TXConstraints() {
    Module             = 0;
    Part               = 0;
    PinPrefix          = "";
    TraceThroughParts  = 0;
    IgnoreThroughParts = 0;
    TraceMaxNetSize    = 10;
    ForceUpperCase     = false;
    GenerateUnique     = false;
    CommentPinName     = false;
    AlignOutput        = false;
    IgnoreGlobal       = false;
    IgnoreNoConnect    = true;
    RemoveGlobalChar   = true;
    UseModulePorts     = true;
    UseDesignNetNames  = true;
  }

  void SetPinPrefix( char* aPinPrefix ) { PinPrefix = aPinPrefix; }

  void AddIgnorePort( char* Port0,
                      char* Port1 = 0, char* Port2 = 0, char* Port3 = 0,
                      char* Port4 = 0, char* Port5 = 0, char* Port6 = 0,
                      char* Port7 = 0, char* Port8 = 0, char* Port9 = 0 );

  void AddIgnorePartPort( char* Port0,
                      char* Port1 = 0, char* Port2 = 0, char* Port3 = 0,
                      char* Port4 = 0, char* Port5 = 0, char* Port6 = 0,
                      char* Port7 = 0, char* Port8 = 0, char* Port9 = 0 );

  void AddIgnoreNet(  char* Port0,
                      char* Port1 = 0, char* Port2 = 0, char* Port3 = 0,
                      char* Port4 = 0, char* Port5 = 0, char* Port6 = 0,
                      char* Port7 = 0, char* Port8 = 0, char* Port9 = 0 );

  void ClearIgnorePorts() {
    IgnorePorts.clear();
  }

  void ClearIgnorePartPorts() {
    IgnorePartPorts.clear();
  }

  void ClearIgnoreNets() {
    IgnorePartPorts.clear();
  }

  void  GenerateConstraintsFile( char*     FileName,                  // destination file name
                                 TModule*  aModule,                   // target module, 0 --> use names from hierarchy's net list
                                 TPart*    aPart );                   // target part
};

#endif

